/*
 * This file is an example of how to embed web-server functionality
 * into existing application.
 * Compilation line:
 * cc example.c shttpd.c -DEMBEDDED -DMT
 */

#ifdef _WIN32
#include <winsock.h>
#define	snprintf		_snprintf
#define	sleep(x)		Sleep((x) * 1000)	
#pragma comment(lib,"ws2_32")
#else
#include <sys/types.h>
#include <sys/select.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <glib.h>
#include <glib/gthread.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include "shttpd.h"

#include "bmpx/dbus.h"
#include "bmpx/metadata.h"

static GMutex *lock;

static gchar* list_tags[] = {
    "artist", 
    "album", 
    "title" 
};

static  DBusGConnection *bus;
static  DBusGProxy	*bmpx;

#define DBUS_TYPE_G_STRING_VALUE_HASHTABLE (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))

static GHashTable *action_method_map = NULL;

static gchar* b_names[] = {
	  "prev",
	  "next",
	  "stop",
	  "play",
	  "pause"
};

static gchar* b_methods[] = {
	  "PlayPrev",
	  "PlayNext",
	  "PlayStop",
	  "PlayCurrent",
	  "PlayPause"
};

#define PAGE_SIZE 50

static int
xmlreply(struct shttpd_callback_arg *arg)
{
  return (0);
}
 
static int
index_html(struct shttpd_callback_arg *arg)
{
    GError *error = NULL;
    gchar **list;
    gint    n = 0;
    gint    c = 0;

    enum {
	N_XHTML,  
	N_BODY,
	N_HEAD,
	N_TABLE,
	N_TR,
	N_TD,
	N_FORM,
	N_BUTTON,
	N_TITLE,
	N_FIELDSET,

	N_NODES
    };

    xmlDocPtr	doc;

    xmlNodePtr	nodes[N_NODES];

    xmlNsPtr	ns;
    xmlDtdPtr	dtd;

    xmlChar    *mem;
    gint	size;

    gint	page,
		length,
		count;

    const gchar  *b;

    g_mutex_lock (lock);

    /* Check controls */
    b = shttpd_get_var (arg->connection, "action"); 
    if (b)
      {
	  if (!dbus_g_proxy_call (bmpx, g_hash_table_lookup (action_method_map, b) , &error, G_TYPE_INVALID, G_TYPE_INVALID))
	    {
		  g_message (error->message);
		  g_error_free (error);
		  error = NULL;
	    }
      }

    b = g_strdup(shttpd_get_var (arg->connection, "page")); 
    if (b)
      page = strtol (b, NULL, 10);
    else
      page = 0;

    /* Output HTTP headers */
    shttpd_printf(arg->connection, "%s\r\n", "HTTP/1.1 200 OK");
    shttpd_printf(arg->connection, "%s\r\n", "Content-Type: text/html");
    shttpd_printf(arg->connection, "%s\r\n", "");

    if (!dbus_g_proxy_call (bmpx, "GetTracklist", &error,
			      G_TYPE_INVALID,
			      G_TYPE_STRV, 
			      &list, 
			      G_TYPE_INVALID))
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetTracklist: %s", error->message);
	return (0);
      }

    length = 0;
    while (list[length]) length++;

    if (!dbus_g_proxy_call (bmpx, "GetCurrentTrack", &error,
			      G_TYPE_INVALID,
			      G_TYPE_INT, 
			      &c, 
			      G_TYPE_INVALID))
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetCurrentTrack: %s", error->message);
	return (0);
      }

      doc = xmlNewDoc(BAD_CAST "1.0");

      nodes[N_XHTML] = xmlNewNode (NULL, BAD_CAST "html"); 
      ns = xmlNewNs (nodes[N_XHTML], BAD_CAST "http://www.w3.org/1999/xhtml", NULL);
      xmlSetNs (nodes[N_XHTML], ns);

      dtd = xmlNewDtd (doc, BAD_CAST "html", BAD_CAST "-//W3C//DTD XHTML 1.0 Strict//EN", BAD_CAST "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd");
      xmlDocSetRootElement (doc, (xmlNodePtr)dtd);

      xmlAddNextSibling ((xmlNodePtr)dtd, nodes[N_XHTML]);	

      nodes[N_HEAD] = xmlNewNode (ns, BAD_CAST "head");

      nodes[N_TITLE] = xmlNewNode (ns, BAD_CAST "title");
      xmlAddChild (nodes[N_TITLE], xmlNewText (BAD_CAST "HTTPDMPX to your service"));    

      nodes[N_BODY] = xmlNewNode (ns, BAD_CAST "body");
      xmlSetProp (nodes[N_BODY], BAD_CAST "style", BAD_CAST "background-color: black; color: orange;");

      xmlAddChild (nodes[N_XHTML], nodes[N_HEAD]);
      xmlAddChild (nodes[N_HEAD], nodes[N_TITLE]);
      xmlAddChild (nodes[N_XHTML], nodes[N_BODY]);

      nodes[N_TABLE] = xmlNewNode (ns, BAD_CAST "table");
      xmlSetProp (nodes[N_TABLE], BAD_CAST "cellspacing", BAD_CAST "4");
      xmlSetProp (nodes[N_TABLE], BAD_CAST "style", BAD_CAST "margin-top: 2em");

      n = 0 + (page * PAGE_SIZE);

      if (n > length)
	  n -= PAGE_SIZE;

      if (n < 0)
	  n  = 0;

      count = 0;
      while ((list[n]) && (count < PAGE_SIZE))
	{
	    GHashTable	*metadata;
	    gchar	*index_str;
	    gint	 m;

	    nodes[N_TR] = xmlNewNode (ns, BAD_CAST "tr");

	    if (!dbus_g_proxy_call (bmpx, "GetMetadataForUri", &error,
			      G_TYPE_STRING,
			      list[n],
			      G_TYPE_INVALID,
			      DBUS_TYPE_G_STRING_VALUE_HASHTABLE, 
			      &metadata, 
			      G_TYPE_INVALID))
	      {
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to complete GetMetadataForUri: %s", error->message);
		return (0);
	      }

	    nodes[N_TD] = xmlNewNode (ns, BAD_CAST "td");

	    if (n == c)
		xmlSetProp (nodes[N_TD], BAD_CAST "style", BAD_CAST "border-left: 6px solid #e73; padding: 0px; padding-left: 10px; padding-right: 4px; text-align: right");
	    else
		xmlSetProp (nodes[N_TD], BAD_CAST "style", BAD_CAST "border-left: 6px solid #999; padding: 0px; padding-left: 10px; padding-right: 4px; text-align: right");

	    index_str = g_strdup_printf ("%d", n+1);

	    xmlAddChild (nodes[N_TD], xmlNewText (BAD_CAST index_str));    
	    g_free (index_str);

	    xmlAddChild (nodes[N_TR], nodes[N_TD]); 

	    for (m = 0; m < 3; m++)
	      {
		  GValue      *value;
		  const gchar *value_str;

		  value = g_hash_table_lookup (metadata, list_tags[m]); 
		  value_str = g_value_get_string (value);

		  nodes[N_TD] = xmlNewNode (ns, BAD_CAST "td");
		  xmlSetProp (nodes[N_TD], BAD_CAST "style", BAD_CAST "padding: 0px; padding-left: 8px; padding-right: 8px");

		  xmlAddChild (nodes[N_TD], xmlNewText (BAD_CAST value_str));    
		  xmlAddChild (nodes[N_TR], nodes[N_TD]); 
	      }

	    xmlAddChild (nodes[N_TABLE], nodes[N_TR]);
	    n++;
	    count++;
	}
      g_free (list);

      // Controls
      nodes[N_FORM] = xmlNewNode  (ns, BAD_CAST "form");
      xmlSetProp (nodes[N_FORM], BAD_CAST "method", BAD_CAST "post");
      xmlSetProp (nodes[N_FORM], BAD_CAST "action", BAD_CAST "/");

      nodes[N_FIELDSET] = xmlNewNode (ns, BAD_CAST "fieldset");
      xmlSetProp (nodes[N_FIELDSET], BAD_CAST "id", BAD_CAST "maincontrols");
      xmlSetProp (nodes[N_FIELDSET], BAD_CAST "title", BAD_CAST "Controls");

      xmlAddChild (nodes[N_FORM], nodes[N_FIELDSET]);
     
      nodes[N_BUTTON] = xmlNewNode  (ns, BAD_CAST "button");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "value", BAD_CAST "prev");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "name", BAD_CAST "action");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "id", BAD_CAST "prev");
      xmlAddChild (nodes[N_BUTTON], xmlNewText (BAD_CAST "Prev"));    
      xmlAddChild (nodes[N_FIELDSET], nodes[N_BUTTON]);

      nodes[N_BUTTON] = xmlNewNode  (ns, BAD_CAST "button");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "value", BAD_CAST "play");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "name", BAD_CAST "action");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "id", BAD_CAST "play");
      xmlAddChild (nodes[N_BUTTON], xmlNewText (BAD_CAST "Play"));    
      xmlAddChild (nodes[N_FIELDSET], nodes[N_BUTTON]);
 
      nodes[N_BUTTON] = xmlNewNode  (ns, BAD_CAST "button");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "value", BAD_CAST "pause");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "name", BAD_CAST "action");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "id", BAD_CAST "pause");
      xmlAddChild (nodes[N_BUTTON], xmlNewText (BAD_CAST "Pause"));    
      xmlAddChild (nodes[N_FIELDSET], nodes[N_BUTTON]);

      nodes[N_BUTTON] = xmlNewNode  (ns, BAD_CAST "button");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "value", BAD_CAST "stop");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "name", BAD_CAST "action");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "id", BAD_CAST "stop");
      xmlAddChild (nodes[N_BUTTON], xmlNewText (BAD_CAST "Stop"));    
      xmlAddChild (nodes[N_FIELDSET], nodes[N_BUTTON]);

      nodes[N_BUTTON] = xmlNewNode  (ns, BAD_CAST "button");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "value", BAD_CAST "next");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "name", BAD_CAST "action");
      xmlSetProp (nodes[N_BUTTON], BAD_CAST "id", BAD_CAST "next");
      xmlAddChild (nodes[N_BUTTON], xmlNewText (BAD_CAST "Next"));    
      xmlAddChild (nodes[N_FIELDSET], nodes[N_BUTTON]);

      xmlAddChild (nodes[N_BODY], nodes[N_FORM]);
      xmlAddChild (nodes[N_BODY], nodes[N_TABLE]);

      xmlKeepBlanksDefault (0);
      xmlThrDefIndentTreeOutput (1);
      xmlDocDumpFormatMemory (doc, &mem, &size, 1);

      shttpd_push (arg->connection, mem, size); 

      g_free (mem);
      xmlFreeDoc (doc);

      g_mutex_unlock (lock);

      return (0);
}

static gpointer
httpd_thread (gpointer data)
{
    gint  
	   sock;

    gint   n;

    /* Initialize and setup URLs we gonna serve */
    shttpd_init(NULL);

    /* Let pass 'data' variable to callback, user can change it */
    shttpd_register_url("/", &index_html, NULL); 
    shttpd_register_url("/xml", &xmlreply, NULL); 

    /* Open two listening sockets */
    sock = shttpd_open_port(9000);

    action_method_map = g_hash_table_new (g_str_hash, g_str_equal);

    for (n = 0; n < 5 ; n++)
	  g_hash_table_insert (action_method_map, g_strdup(b_names[n]), g_strdup(b_methods[n]));

    while (1)
      {
#if 0
          FD_ZERO(&rfds);
	  FD_SET(sock, &rfds);

	  shttpd_merge_fds(&rfds, NULL, &sock);

	  rv = select(sock+1, &rfds, NULL, NULL, NULL);
#endif

	  shttpd_poll (sock, 100);
      }

    /* Probably unreached, because we will be killed by signal */
    shttpd_fini();
}

gint
main(gint argc, gchar **argv)
{
	GError	    *error = NULL;
	GMainLoop   *mainloop;
	GThread	    *thread;

	/* Initialize GLib */
	g_type_init ();
	g_thread_init (NULL); 
	dbus_g_type_specialized_init ();

	bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (!bus)
	  {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Couldn't connect to session bus: %s", error->message);
	      exit (EXIT_FAILURE);
	  }

	lock = g_mutex_new ();

	bmpx = dbus_g_proxy_new_for_name (bus,
					     BMP_DBUS_SERVICE,
					     BMP_DBUS_PATH_SYSTEMCONTROL,
					     BMP_DBUS_INTERFACE);


 
	mainloop = g_main_loop_new (NULL, FALSE);

	thread = g_thread_create (httpd_thread,
				  NULL,
				  TRUE,
				  NULL);

	g_main_loop_run (mainloop);
	
	return (EXIT_SUCCESS);
}
