#include <dbus/dbus-glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bmpx/dbus.h"

int
main (int argc, char **argv)
{
  DBusGConnection   *bus;
  DBusGProxy	    *remote_object;
  GError	    *error = NULL;
  gchar		   **array_list;
  gchar		    *aux;
  gint		     n = 0;
  gchar		    *filename;

  g_type_init ();
  dbus_g_type_specialized_init ();

  {
    GLogLevelFlags fatal_mask;
    
    fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
    fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
    g_log_set_always_fatal (fatal_mask);
  }

  bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (!bus)
    {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Couldn't connect to session bus: %s", error->message);
	exit (EXIT_FAILURE);
    }
  
  remote_object = dbus_g_proxy_new_for_name (bus,
					     BMP_DBUS_SERVICE,
					     BMP_DBUS_PATH_SYSTEMCONTROL,
					     BMP_DBUS_INTERFACE);

  printf ("<?xml version='1.0'?>\n");
  printf("<bmp:playlist>\n");

  if (!dbus_g_proxy_call (remote_object, "GetTracklist", &error,
			  G_TYPE_INVALID,
			  G_TYPE_STRV, 
			  &array_list, 
			  G_TYPE_INVALID))
    {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Failed to complete GetTracklist: %s", error->message);
    }
	
  for ( ; *array_list; array_list++)
    {
		filename = *array_list; 
		aux = g_markup_escape_text(filename, strlen(filename));
		printf ("\t<bmp:playlist-file position='%d'>%s</playlist-file>\n", n++, aux);
		g_free(aux);
    }

  printf("\n</bmp:playlist>\n\n");
 
  exit(EXIT_SUCCESS);
}
