#
# KBMPx gdesklets Control
# (C) 2006 Milosz Derezynski
# version 0.1

from libdesklets.controls import Control

import dbus 
import dbus.glib
import string
import os
import time

from IKBMPx import IKBMPx

class KBMPx(Control, IKBMPx):
	def __init__(self):
		Control.__init__(self)

		self._title	= ""
		self._artist	= ""
		self._ctime	= 0
		self._cpercent	= 0
		self._p_length	= 0
		self._p_uri	= ""

		bus	  = dbus.SessionBus()
		obj	  = bus.get_object('org.mpris.bmpx', '/org/mpris/bmpx/SystemControl')
		self.bmpx = dbus.Interface (obj, 'org.mpris.bmpx')

		self.bmpx.connect_to_signal('TrackChange', self.bmpx_track_change)
		self.bmpx.connect_to_signal('SetTitle', self.bmpx_set_title)
		self.bmpx.connect_to_signal('SetStreamPos', self.bmpx_set_stream_pos)
  
		self.bmpx.SendStatus()

	# DBus signal handlers
	def bmpx_track_change(self):
		self._p_uri = self.bmpx.GetCurrentUri()
		metadata = self.bmpx.GetMetadataForUri(self._p_uri)

		self._p_length = int(metadata["time"])

		self._title = metadata["title"]
		self._artist = metadata["artist"]
		self._update("title")
		self._update("artist")

	def bmpx_set_title(self):
		self._p_uri = self.bmpx.GetCurrentUri()
		metadata = self.bmpx.GetMetadataForUri(self._p_uri)

		self._title = metadata["title"]
		self._artist = metadata["artist"]
		self._update("title")
		self._update("artist")

	def bmpx_set_stream_pos (self, stream_pos):
		percent = self._p_length / 100;
		self._ctime = stream_pos
		self._cpercent = stream_pos / percent;
		self._update("ctime")
		self._update("cpercent")

	def __title(self):
		return self._title

	def __artist(self):
		return self._artist

	def __ctime(self):
		return self._ctime

	def __cpercent(self):
		return self._cpercent

	def __action(self, action):

                if (action == "stop"):
                        self.bmpx.PlayStop()

                elif (action == "play"):
                        self.bmpx.Play()

                elif (action == "next"):
                        self.bmpx.PlayNext()

                elif (action == "prev"):
                        self.bmpx.PlayPrev()

                elif (action == "pause"):
                        self.bmpx.PlayPause()

	artist	  = property(fget = __artist)	
	title	  = property(fget = __title)	
	ctime	  = property(fget = __ctime)
	cpercent  = property(fget = __cpercent)
	action	  = property(fset = __action)

def get_class(): return KBMPx
