/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <glib-object.h>
#include <gtk/gtk.h>

#include <bmpx/paramspecs.h>
#include <bmpx/plugin.h>
#include <bmpx/playlist.h>
#include <bmpx/metadata.h>
#include <bmpx/plugin_interfaces.h>

#include "libflow_linear.h"

static gint datum_id[] = {
  BMP_DATUM_ARTIST,
  BMP_DATUM_ALBUM,
  BMP_DATUM_TITLE
};

typedef struct _Options Options;
struct _Options {
    GList    *specs;
  
    GQuark    use_match_q;
    gboolean  use_match;

    GQuark    match_q;
    gchar    *match;

    GQuark    list_q;
    gint      list;
};

#define _OPTION_USE_MATCH "use-match"
#define _OPTION_MATCH     "match"
#define _OPTION_LIST      "list"

static Options   *options = NULL;

/* Linear Flow Control */
static gboolean 
flow_linear_item_next (BmpPluginFlowInterface	  *self, 
		       GtkListStore		  *playlist,
		       GtkTreeIter		  *iter) 
{
    BmpFileItem		   *file_item;
    GtkTreeRowReference	   *reference;
    GtkTreePath		   *path;
    gint		    length;
    gint		    current;

    length = gtk_tree_model_iter_n_children (GTK_TREE_MODEL(playlist), NULL)-1; 
    if (length <= 0)
    {
	return FALSE;
    }

    reference = g_object_get_data (G_OBJECT(playlist), "current");
    if (!reference)
    {
	gtk_tree_model_get_iter_first (GTK_TREE_MODEL(playlist), iter);
	return TRUE;
    }

    path = gtk_tree_row_reference_get_path (reference);
    if (!path)
    {
	return FALSE;
    }


    current = gtk_tree_path_get_indices (path)[0]; 
    gtk_tree_path_free (path);

    if ((current+1) > length)
	{
	    return FALSE;
	}
  
    retry_next:
    current++;

    if (current <= length)
      {
	gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(playlist), iter, NULL, current);

	if ((options->use_match) && (options->match) && (strlen(options->match)))
	  {
	      HTupleElem *tuple_elem;
	      gchar	 *a, *b;

	      gtk_tree_model_get (GTK_TREE_MODEL(playlist), iter, 5, &file_item, -1);
	      tuple_elem = h_tuple_get (file_item->tuple, bmp_metadata_get_id_static (datum_id[options->list])); 
  
	      a = g_utf8_strdown ((gchar*)tuple_elem->value, strlen((gchar*)tuple_elem->value));
	      b = g_utf8_strdown ((gchar*)options->match, strlen((gchar*)options->match));

	      if (!g_strstr_len (a, strlen (a), b))
		{
		    goto retry_next;
		}
	  }

	  return TRUE;
      }

  return FALSE;
}

static gboolean 
flow_linear_item_prev (BmpPluginFlowInterface	    *self,
		       GtkListStore		    *playlist,
		       GtkTreeIter		    *iter) 
{
    BmpFileItem		   *file_item;
    GtkTreeRowReference	   *reference;
    GtkTreePath		   *path;
    gint		    length;
    gint		    current;

    length = gtk_tree_model_iter_n_children (GTK_TREE_MODEL(playlist), NULL)-1; 
    reference = g_object_get_data (G_OBJECT(playlist), "current");
    if (!reference) return FALSE; 

    path = gtk_tree_row_reference_get_path (reference);
    if (!path) return FALSE;

    current = gtk_tree_path_get_indices (path)[0]; 
    gtk_tree_path_free (path);

    retry_prev:
    current--;

    if (current >= 0)
      {
	gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(playlist), iter, NULL, current);

	if ((options->use_match) && (options->match) && (strlen(options->match)))
	  {
	      HTupleElem *tuple_elem;
	      gchar	 *a, *b;

	      gtk_tree_model_get (GTK_TREE_MODEL(playlist), iter, 5, &file_item, -1);
	      tuple_elem = h_tuple_get (file_item->tuple, bmp_metadata_get_id_static (datum_id[options->list])); 
  
	      a = g_utf8_strdown ((gchar*)tuple_elem->value, strlen((gchar*)tuple_elem->value));
	      b = g_utf8_strdown ((gchar*)options->match, strlen((gchar*)options->match));

	      if (!g_strstr_len (a, strlen (a), b))
		{
		    goto retry_prev;
		}
	  }

	  return TRUE;
      }

    return FALSE;
}

static GList*
prop_get_props (BmpPlugPropertyInterface *self)
{
	return options->specs;
}

static gboolean
prop_set_prop  (BmpPlugPropertyInterface *self,
		GQuark		          prop,
	        GValue			 *value)
{
    if (prop == options->use_match_q)
      {
	  options->use_match = g_value_get_boolean (value);
	  return TRUE;
      }

    if (prop == options->list_q)
      {
	  options->list = g_value_get_int (value);
	  return TRUE;
      }

     if (prop == options->match_q)
      {
	  if (options->match) g_free (options->match);
	  options->match = g_value_dup_string (value);
	  return TRUE;
      }
 
    return FALSE;
}

static void
prop_init (gpointer         g_iface,
           gpointer         iface_data)
{
        BmpPlugPropertyInterface *klass = (BmpPlugPropertyInterface *)g_iface;
	GList			 *list = NULL;

        klass->get_props = prop_get_props;
	klass->set_prop  = prop_set_prop;

	options = g_new0 (Options,1);
	options->specs = NULL;

	/* Use match */ 
	options->specs = g_list_append (options->specs,
				        g_param_spec_boolean (_OPTION_USE_MATCH,
							      _OPTION_USE_MATCH,
							      "Use matching",
							       FALSE, BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->use_match = FALSE;
	options->use_match_q = g_quark_from_static_string (_OPTION_USE_MATCH);

	/* Match */
	options->specs = g_list_append (options->specs,
				        g_param_spec_string  (_OPTION_MATCH,
							      _OPTION_MATCH,
							      "Match text",
							       "", BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->match = g_strdup("");
	options->match_q = g_quark_from_static_string (_OPTION_MATCH);

	/* List */
	list = g_list_append (list, bmp_metadata_get_title (datum_id[0]));
	list = g_list_append (list, bmp_metadata_get_title (datum_id[1]));
	list = g_list_append (list, bmp_metadata_get_title (datum_id[2]));

	options->specs = g_list_append (options->specs,
				        bmp_param_spec_list  (_OPTION_LIST,
							      _OPTION_LIST,
							      "Match Metadatum",
							       list, 0, BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->list = 0; 
	options->list_q = g_quark_from_static_string (_OPTION_LIST);
}


static gchar* _metadata[] = {
    "Linear Playback",
    "Milosz Derezynski",
    "(C) GPL 2005",
    "http://bmpx.beep-media-player.org",
    "internalerror@gmail.com",
};

const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
    return (const gchar**)_metadata;
}

static void
metadata_init (gpointer         g_iface,
               gpointer         iface_data)
{
        BmpPlugMetadataInterface *klass = (BmpPlugMetadataInterface *)g_iface;

        klass->get_metadata = get_metadata;
}


static void
flow_linear_interface_init (gpointer         g_iface,
                            gpointer         iface_data)
{
        BmpPluginFlowInterface *klass = (BmpPluginFlowInterface *)g_iface;

        klass->item_next   = flow_linear_item_next;
        klass->item_prev   = flow_linear_item_prev;
        klass->reset_state = NULL; 
}

GType plugin_flow_linear_type = 0;

void 
plugin_flow_linear_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginFlowLinearClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginFlowLinear),
                        0,      /* n_preallocs */
                        NULL,    /* instance_init */
                };

                plugin_flow_linear_type = g_type_module_register_type (module,
					   G_TYPE_OBJECT,
					  "BmpPluginFlowLinearType",
					   &object_info, 0);

                static const GInterfaceInfo interface_flow_linear_info = {
                        (GInterfaceInitFunc) flow_linear_interface_init, 
                        NULL,                                    
                        NULL                                      
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    
                        NULL,                                 
                        NULL                                 
                };

                static const GInterfaceInfo interface_property_info = {
                        (GInterfaceInitFunc) prop_init,		   
                        NULL,                                     
                        NULL                                     
                };

                g_type_module_add_interface (module,
					     plugin_flow_linear_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);

                g_type_module_add_interface (module,
					     plugin_flow_linear_type,
                                             BMP_TYPE_PLUGIN_FLOW,
                                             &interface_flow_linear_info);

               g_type_module_add_interface (module,
					     plugin_flow_linear_type,
                                             BMP_TYPE_PLUG_PROPERTY,
                                             &interface_property_info);

}

