/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <config.h>

#include <string.h>
#include <assert.h>
#include <signal.h>
#include <getopt.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <glib.h>
#include <glib-object.h>

#include <config.h>

#include <bmpx/util.h>
#include <bmpx/uri.h>
#include <bmpx/metadata.h>

#include <xcs/xcs.h>
#include <goa/libgoa.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>

#include <xine.h>

#include "play.h"
#include "main.h"

#include "bmpx-marshalers.h"

#define BMP_PLAY_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), BMP_TYPE_PLAY, BmpPlayPrivate))

G_DEFINE_TYPE (BmpPlay, bmp_play, G_TYPE_OBJECT)

G_LOCK_DEFINE (METADATA_XINE_LOCK);

enum
{
  BMP_PLAY_SIGNAL_EOS,
  BMP_PLAY_SIGNAL_ERROR,
  BMP_PLAY_SIGNAL_BUFFER,
  BMP_PLAY_SIGNAL_TIME_TICK,
  BMP_PLAY_SIGNAL_SET_STREAM_TITLE,

  BMP_PLAY_N_SIGNALS
};

static guint signals[BMP_PLAY_N_SIGNALS] = { 0 };

enum
{
  BMP_PLAY_PROP_STREAM = 1,
  BMP_PLAY_PROP_PLAYSTATUS = 2,
  BMP_PLAY_N_PROPS
};

typedef struct _BmpPlayPrivate BmpPlayPrivate;

struct _BmpPlayPrivate
{
  gboolean dispose_has_run;

  xine_t             *xine;
  xine_stream_t      *stream;
  xine_audio_port_t  *audioport;
  xine_post_t        *post;
  xine_event_queue_t *evq;

  gchar              *mrl;
  gint                playstatus;

  gint64              current_vpts;

  guint		      source;
  gboolean	      sending;
};

static GObject *bmp_play_constructor (GType type,
                                      guint n_construct_properties,
                                      GObjectConstructParam *
                                      construct_properties);
static void
send_stream_pos_stop  (BmpPlay *self);
 
static void
send_stream_params    (BmpPlay * self);

#if 0
static void
prune_buffers (BmpPlay *self);
#endif

static gboolean
_idle_emit_eos (BmpPlay * self)
{
  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  g_signal_emit (G_OBJECT (self), signals[BMP_PLAY_SIGNAL_EOS], 0, NULL);

  return FALSE;
}

#if 0
static gboolean
_fadeout (BmpPlay * self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint            volume;

  volume = xine_get_param (private->stream, XINE_PARAM_AUDIO_VOLUME);
  volume -= 10;

  if (volume <= 0)
    {
      xine_stop (private->stream);
      xine_close (private->stream);
      send_stream_params (self);
      g_object_set (G_OBJECT (self), "playstatus", PLAYSTATUS_STOPPED,
                    NULL);
      return FALSE;
    }
  else
    {
      xine_set_param (private->stream, XINE_PARAM_AUDIO_VOLUME,
                      volume);
      return TRUE;
    }
}
#endif

void
bmp_play_set_volume (BmpPlay *self,
                     gint     volume)
{
  BmpPlayPrivate *private;

  g_return_if_fail (BMP_IS_PLAY (self));

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return;

  xine_set_param (private->stream, XINE_PARAM_AUDIO_VOLUME, volume);
}

void
bmp_play_set_eq (BmpPlay   *self,
                 BmpEqEnum  eq,
                 gint       value)
{
  BmpPlayPrivate *private;

  g_return_if_fail (BMP_IS_PLAY (self));

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return;

  xine_set_param (private->stream, eq + BMP_EQ_TO_ENGINE_OFFSET,
                  value);
}

#if 0
static void
playobjects_destroy (BmpPlay * self)
{
  g_return_if_fail (BMP_IS_PLAY (self));
}
#endif

static void
event_listener_cb (gpointer user_data, const xine_event_t * event)
{
  switch (event->type)
  {
  case XINE_EVENT_UI_PLAYBACK_FINISHED:
    {
      g_idle_add ((GSourceFunc) _idle_emit_eos, user_data);
    }
    break;

  case XINE_EVENT_UI_SET_TITLE:
    {
      xine_ui_data_t *ui_data = event->data;

      g_signal_emit (G_OBJECT (user_data), signals[BMP_PLAY_SIGNAL_SET_STREAM_TITLE], 0, g_strdup(ui_data->str));
    }
    break;

  default:;
  }
}

static void
playobjects_create (BmpPlay *self)
{
#if 0
  xine_post_out_t *source;
  xine_post_in_t  *target;
#endif

  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gchar          *filename;

  private->xine = xine_new ();

  filename = g_build_filename (g_get_home_dir (), ".xine", "config", NULL);
  xine_config_load (private->xine, filename);
  g_free (filename);
  xine_init (private->xine);

  private->audioport = xine_open_audio_driver (private->xine, NULL, NULL);
  private->stream = xine_stream_new (private->xine, private->audioport, NULL);
  private->evq = xine_event_new_queue (private->stream);
  xine_event_create_listener_thread (private->evq, event_listener_cb, self);

#if 0
  private->post = scope_plugin_new ( private->xine,
                                           private->audioport);
#endif

  xine_set_param( private->stream, XINE_PARAM_METRONOM_PREBUFFER, 6000 );
  xine_set_param( private->stream, XINE_PARAM_IGNORE_VIDEO, 1 );

#if 0
  source = xine_get_audio_source ( private->stream );
  target = (xine_post_in_t*)xine_post_input( private->post , (const gchar*)("audio in") );
  xine_post_wire ( source, target );
#endif
}

gboolean
bmp_play_stop_stream (BmpPlay *self)
{
  BmpPlayPrivate *private;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return TRUE;

  xine_stop (private->stream);
  xine_close (private->stream);
  send_stream_params (self);
  send_stream_pos_stop (self);

#if 0
  prune_buffers (self);
#endif

  return TRUE;
}

#if 0
gboolean
bmp_play_stop_stream (BmpPlay *self)
{
  BmpPlayPrivate *private;
  gint            status;


  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return TRUE;

  status = xine_get_status (private->stream);

  if (status != XINE_STATUS_PLAY)
    return TRUE;

  g_timeout_add (75, _fadeout, self);

  return TRUE;
}
#endif

static void
send_stream_params (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint            status;
  guint32         bitrate, samplerate;

  if (!private->stream)
    return;

  status = xine_get_status (private->stream);

  if (status != XINE_STATUS_STOP)
    {
      bitrate = xine_get_stream_info (private->stream,
                                      XINE_STREAM_INFO_AUDIO_BITRATE);
      bmp_system_control_bitrate (bmp_system_control, bitrate);

      samplerate = xine_get_stream_info (private->stream,
                                         XINE_STREAM_INFO_AUDIO_SAMPLERATE);
      bmp_system_control_samplerate (bmp_system_control, samplerate);
    }
  else
    {
      bmp_system_control_bitrate (bmp_system_control, 0);
      bmp_system_control_samplerate (bmp_system_control, 0);
    }
}

#if 0
static void
prune_buffers (BmpPlay *self)
{
  // here we prune the buffer list regularly

  MyNode *myList = scope_plugin_list( private->post );
  MyNode *prev, *node;

  // we operate on a subset of the list for thread-safety
  MyNode * const first_node = myList->next;
  MyNode const * const list_end = myList;

  private->current_vpts =
    (xine_get_status (private->stream) == XINE_STATUS_PLAY) ?
      xine_get_current_vpts (private->stream) : G_MAXINT64; // if state is not playing OR paused, empty the list

  node = first_node->next;

  for (prev = first_node; node != list_end; node = node->next)
    {
      // we never delete first_node
      // this maintains thread-safety
      if (node->vpts_end < private->current_vpts)
        {
          prev->next = node->next;

          free(node->mem);
          free(node);

          node = prev;
        }

      prev = node;
    }
}
#endif

#if 0
typedef gint16* Scope;

Scope
process_scope (BmpPlay *self)
{
  BmpPlayPrivate *private;

  g_return_val_if_fail (BMP_IS_PLAY (self), NULL);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->post || xine_get_status (private->stream) != XINE_STATUS_PLAY)
    return NULL;

  MyNode* const myList         = scope_plugin_list( private->post );
  metronom_t* const myMetronom = scope_plugin_metronom( private->post );
  const int myChannels         = scope_plugin_channels( private->post );
  int n = 0, frame = 0;
  Scope m_scope;

  // prune the buffer list and update private->current_vpts
  prune_buffers (self);

  m_scope = g_new0 (gint16, 512);

  for (n = 0; frame < 512; )
    {
      MyNode *best_node = 0;

      for (MyNode *node = myList->next; node != myList; node = node->next)
        if (node->vpts <= private->current_vpts && (!best_node || node->vpts > best_node->vpts))
          best_node = node;

      if (!best_node || best_node->vpts_end < private->current_vpts)
        break;

      gint64 diff = private->current_vpts;
      diff -= best_node->vpts;
      diff *= 1<<16;
      diff /= myMetronom->pts_per_smpls;

      const gint16 *data16 = best_node->mem;
      data16 += diff;

      diff += diff % myChannels; // important correction to ensure we don't overflow the buffer
      diff /= myChannels;        // use units of frames, not samples

      // calculate the number of available samples in this buffer
      n  = best_node->num_frames;
      n -= diff;
      n += frame; //clipping for # of frames we need

      if (n > 512)
        n = 512; //we don't want more than 512 frames

      gint a, c;

      for (; frame < n; ++frame, data16 += myChannels )
        {
          for( a = c = 0; c < myChannels; ++c )
            a += data16[c];

          a /= myChannels;
          m_scope[frame] = a;
        }

      private->current_vpts = best_node->vpts_end;
      private->current_vpts++; // FIXME needs to be done for some reason, or you get situations where it uses same buffer again and again
    }

  return m_scope;
}
#endif

#if 0
static gboolean
upload_audio_data (BmpPlay *self)
{
  gint stream_pos = bmp_play_get_time (self);
  guint32 status;

  if (!private->stream)
    return FALSE;

  status = xine_get_status (private->stream);

  if (status != XINE_STATUS_PLAY)
    return FALSE;

  if ((stream_pos >= 0) && (status == XINE_STATUS_PLAY))
    {
      Scope m_scope;
      gint16 data[2][512];

      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", stream_pos, NULL);

      m_scope = process_scope (self);
      if (m_scope)
        {
          memcpy (data[0], m_scope, 512);
          memcpy (data[1], m_scope, 512);
          lv_bmp_render_pcm (data);
          g_free (m_scope);
        }
    }

  return TRUE;
}
#endif

static void
send_stream_pos_stop (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GSource *source;

  if (private->source)
    {
	source = g_main_context_find_source_by_id (NULL, private->source);
	if (source)
	  {
	    g_source_destroy (source);
	    private->sending = FALSE;
	  }
    }
}

static gboolean
send_stream_pos (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint            stream_pos = bmp_play_get_time (self);
  guint32         status;

  if (!private->stream)
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      return FALSE;
    }

  status = xine_get_status (private->stream);

  if (status != XINE_STATUS_PLAY)
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      g_object_set (G_OBJECT (self), "playstatus", PLAYSTATUS_STOPPED, NULL);
      return FALSE;
    }

  if ((stream_pos >= 0) && (status == XINE_STATUS_PLAY))
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", stream_pos, NULL);
      return TRUE;
    }
  else
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      return FALSE;
    }

  return TRUE;
}

gchar *
bmp_play_get_suffixes (BmpPlay *self)
{
  g_return_val_if_fail (BMP_IS_PLAY (self), NULL);

  return xine_get_file_extensions (BMP_PLAY_GET_PRIVATE (self)->xine);
}

gchar **
bmp_play_get_cd (BmpPlay  *self,
                 gint     *n_mrls,
                 gchar   **disc_id)
{
  BmpPlayPrivate *private;
  xine_stream_t  *stream_aux;
  gchar         **mrls;

  g_return_val_if_fail (BMP_IS_PLAY (self), NULL);

  private = BMP_PLAY_GET_PRIVATE (self);

  mrls = xine_get_autoplay_mrls (private->xine, "CD", n_mrls);
  if (!mrls) return NULL;

  stream_aux = xine_stream_new (private->xine, NULL, NULL);
  xine_open (stream_aux, mrls[0]);

  *disc_id = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_CDINDEX_DISCID));

  xine_close (stream_aux);
  xine_dispose (stream_aux);

  return mrls;
}

gboolean
bmp_play_play_stream (BmpPlay * self)
{
  BmpPlayPrivate *private;
  gchar          *mrl_local;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (xine_get_param (private->stream, XINE_PARAM_SPEED) == XINE_SPEED_PAUSE)
    {
      xine_set_param (private->stream, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
      return TRUE;
    }

  xine_stop (private->stream);
  xine_close (private->stream);

  if (!g_ascii_strncasecmp (private->mrl, "cdda", 4))
    {
      mrl_local =
      get_cdda_mrl_from_uri (private->mrl);
    }
  else
    {
      mrl_local =
      g_strdup (private->mrl);
    }

  if (!xine_open (private->stream, mrl_local))
    {
      g_message ("%s: %d", G_STRLOC,
                 xine_get_error (private->stream));
      bmp_system_message_send (bmp_system_control,
                               BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
      g_free (mrl_local);
      return TRUE;
    }

  if (!xine_play (private->stream, 0, 0))
    {
      g_message ("%s: %d", G_STRLOC,
                 xine_get_error (private->stream));
      bmp_system_message_send (bmp_system_control,
                               BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
      g_free (mrl_local);
      return TRUE;
    }

  bmp_play_set_volume (self,
                       g_object_get_int (bmp_system_control, "volume"));

  send_stream_params (self);

  if (!private->sending)
    {
      private->source = g_timeout_add_full (G_PRIORITY_DEFAULT, 250,
                      (GSourceFunc) send_stream_pos, self, NULL);
      private->sending = TRUE;
    }

#if 0
  g_timeout_add_full (G_PRIORITY_DEFAULT, 25,
                      (GSourceFunc) upload_audio_data, self, NULL);
#endif

  g_free (mrl_local);
  return TRUE;
}

struct meta_thread_data_t
{
  GMutex  *mutex;
  BmpPlay *self;
  gchar   *uri;
  HTuple  *tuple;
};
#define META_THREAD_DATA(_data) ((struct meta_thread_data_t *)(_data))

static GCond    *metadata_cond = NULL;

static gpointer
meta_fetch_thread_xine (gpointer data)
{
  BmpPlay        *self = META_THREAD_DATA (data)->self;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  gchar          *_uri = META_THREAD_DATA (data)->uri;

  GURI           *bmp_uri;

  HTuple         *tuple;

  xine_stream_t  *stream_aux;

  gint            stream_pos, stime, length;

  const gchar    *aux;

  gchar          *uri, 
		 *aux_check,
		 *uri_str;

  struct stat	  file_stat;

  bmp_uri = bmp_uri_new (_uri);
  bmp_uri_unescape (bmp_uri);

  uri_str = bmp_uri_get_string (bmp_uri);
  tuple = bmp_metadata_tuple_create_default (uri_str);
  g_free (uri_str);

  stat (bmp_uri->path, &file_stat);

  h_tuple_insert (tuple, 
		  bmp_metadata_get_id_static (BMP_DATUM_MTIME),    
		  G_TYPE_LONG,    
		  (gpointer)(file_stat.st_mtime));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_LOCATION),
                  G_TYPE_STRING,
                  (gpointer)_uri);


  if (!strcmp("cdda",bmp_uri->scheme))
    uri = get_cdda_mrl_from_uri (_uri);
  else
    uri = g_strdup (_uri);

  stream_aux = xine_stream_new (private->xine, NULL, NULL);

  if (!xine_open (stream_aux, uri))
    {
      g_message ("%s: could not open stream '%s'", G_STRLOC, uri);
      g_free (uri);
      xine_dispose (stream_aux);

      /* If we cannot connect, return default values, else the playlist
       * expose code might stall if it tries to connect again. */
      META_THREAD_DATA (data)->tuple = tuple;

      /* we are done, so transport_http_handle_read() can return the data */
      g_mutex_lock (META_THREAD_DATA(data)->mutex);
      g_cond_signal (metadata_cond);
      g_mutex_unlock (META_THREAD_DATA (data)->mutex);

      g_thread_exit (NULL);

      /* shutup gcc */
      return NULL;
    }

  xine_get_pos_length (stream_aux, &stream_pos, &stime, &length);
  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_TRACK_NUMBER);

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_TRACK),
                  G_TYPE_INT,
                  GINT_TO_POINTER(aux ? strtol (aux, NULL, 10) : 0));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_DATE),
                  G_TYPE_INT,
                  GINT_TO_POINTER(0));

  /* G_TYPE_STRING metadata */
  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_ARTIST);
  aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL;
  if (aux && strlen(aux_check))
    {
      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_ARTIST),
                      G_TYPE_STRING,
                      (gpointer)aux);
    }
  g_free (aux_check);

  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_ALBUM);
  aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL;
  if (aux && strlen(aux_check))
    {
      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_ALBUM),
                      G_TYPE_STRING,
                      (gpointer)aux);
    }
  g_free (aux_check);

  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_TITLE);
  aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL;
  if (aux && strlen(aux_check))
    {
      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_TITLE),
                      G_TYPE_STRING,
                      (gpointer)aux);
    }
  else
    {
      gchar  *basename;
      gchar **fragments;

      basename = g_path_get_basename (bmp_uri->path); 
      fragments = g_strsplit (basename, ".", -1);

      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_TITLE),
                      G_TYPE_STRING,
                      (gpointer)fragments[0]);
      g_strfreev (fragments);
      g_free (basename);
    }
  g_free (aux_check);

  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_GENRE);
  aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL;
  if (aux && strlen (aux_check))
    {
      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_GENRE),
                      G_TYPE_STRING,
                      (gpointer)aux);
    }
  g_free (aux_check);

  aux = xine_get_meta_info (stream_aux, XINE_META_INFO_COMMENT);
  aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL;
  if (aux && strlen(aux_check))
    {
      h_tuple_insert (tuple,
                      bmp_metadata_get_id_static (BMP_DATUM_COMMENT),
                      G_TYPE_STRING,
                      (gpointer)aux);
    }
  g_free (aux_check);


  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_BITRATE),
                  G_TYPE_INT,
                  GINT_TO_POINTER(xine_get_stream_info (stream_aux, XINE_STREAM_INFO_AUDIO_BITRATE)/1000));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),
                  G_TYPE_INT,
                  GINT_TO_POINTER(xine_get_stream_info (stream_aux, XINE_STREAM_INFO_AUDIO_SAMPLERATE)));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_TIME),
                  G_TYPE_INT,
                  GINT_TO_POINTER(length/1000));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_RATING),
                  G_TYPE_INT,
                  GINT_TO_POINTER(0));

  h_tuple_insert (tuple,
                  bmp_metadata_get_id_static (BMP_DATUM_COUNT),
                  G_TYPE_INT,
                  GINT_TO_POINTER(0));

  g_free (uri);
  xine_close (stream_aux);
  xine_dispose (stream_aux);

  META_THREAD_DATA (data)->tuple = tuple;

  /* we are done, so transport_http_handle_read() can return the data */

  g_mutex_lock (META_THREAD_DATA(data)->mutex);
  g_cond_signal (metadata_cond);
  g_mutex_unlock (META_THREAD_DATA (data)->mutex);

  g_thread_exit (NULL);

  /* shutup gcc */
  return NULL;
}

HTuple *
bmp_play_get_metadata_xine (BmpPlay * self, const gchar *_uri)
{
  GThread *meta_thread;
  GTimeVal _tv;
  struct meta_thread_data_t thread_data;

  g_return_val_if_fail (BMP_IS_PLAY(self), NULL);

  G_LOCK (METADATA_XINE_LOCK);

  thread_data.self = self;
  thread_data.uri = (gchar *)_uri;
  thread_data.tuple = NULL;

  metadata_cond = g_cond_new ();

  /* get a new mutex, and lock it - the thread will unlock it when done */
  thread_data.mutex = g_mutex_new ();
  g_mutex_lock (thread_data.mutex);

  /* setup data to be passed to thread, and create new thread to handle
   * fetching the http data */
  meta_thread = g_thread_create (meta_fetch_thread_xine,
                                 &thread_data,
                                 FALSE,
                                 NULL);

  /* wait while we busy fetching the data and make sure signals are passed */
  while (42)
    {
      while (gtk_events_pending ())
          gtk_main_iteration ();

      g_get_current_time (&_tv);
      g_time_val_add (&_tv, 10000);

      if (g_cond_timed_wait(metadata_cond, thread_data.mutex, &_tv))
        break;
    }

  /* all done, unlock and free mutex */
  g_mutex_unlock (thread_data.mutex);
  g_mutex_free (thread_data.mutex);

  G_UNLOCK (METADATA_XINE_LOCK);

  return thread_data.tuple;
}

gboolean
bmp_play_pause_stream (BmpPlay *self)
{
  BmpPlayPrivate *private;
  gint            speed;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  speed = xine_get_param (private->stream, XINE_PARAM_SPEED);

  switch (speed)
    {
    case XINE_SPEED_PAUSE:
      g_object_set (self, "playstatus", PLAYSTATUS_PLAYING, NULL);
      break;

    case XINE_SPEED_NORMAL:
      xine_set_param (private->stream, XINE_PARAM_SPEED,
                      XINE_SPEED_PAUSE);
      break;
    }

  return TRUE;
}

gboolean
bmp_play_seek_stream (BmpPlay *self,
                      gint     offset)
{
  BmpPlayPrivate *private;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return FALSE;

  return !xine_play (private->stream, 0, offset) ? TRUE : FALSE;
}

gint
bmp_play_get_status (BmpPlay *self)
{
  BmpPlayPrivate *private;
  gint            status;

  g_return_val_if_fail (BMP_IS_PLAY (self), PLAYSTATUS_STOPPED);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return FALSE;

  status = xine_get_status (private->stream);

  switch (status)
    {
    case XINE_STATUS_STOP:
      return PLAYSTATUS_STOPPED;

    case XINE_STATUS_PLAY:
      return PLAYSTATUS_PLAYING;

    default:
      return status;
    }
}

gint
bmp_play_get_time (BmpPlay *self)
{
  BmpPlayPrivate *private;
  gint            stream_pos, length, stime;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return 0;

  xine_get_pos_length (private->stream, &stream_pos, &stime, &length);
  return stime / 1000;
}

gint
bmp_play_get_length (BmpPlay *self)
{
  BmpPlayPrivate *private;
  gint            stream_pos, length, stime;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->stream)
    return 0;

  xine_get_pos_length (private->stream, &stream_pos, &stime, &length);

  return length / 1000;
}

static gboolean
playstatus_notify_cb (BmpPlay *self)
{
  gint status;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

  g_object_get (self, "playstatus", &status, NULL);

  if (status & PLAYSTATUS_SEEKING)
    {
      bmp_play_pause_stream (self);
      return TRUE;
    }

  if (status & PLAYSTATUS_STOPPED)
    {
      bmp_play_stop_stream (self);
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      return TRUE;
    }

  if (status & PLAYSTATUS_WAITING)
    {
      bmp_play_stop_stream (self);
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      return TRUE;
    }

  if (status & PLAYSTATUS_PAUSED)
    {
      bmp_play_pause_stream (self);
      return TRUE;
    }

  if (status & PLAYSTATUS_PLAYING)
    {
      bmp_play_play_stream (self);
      return TRUE;
    }

  return FALSE;
}

static void
bmp_play_set_property (GObject * object,
                       guint property_id,
                       const GValue * value, GParamSpec * pspec)
{
  BmpPlay        *self    = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
    {
    case BMP_PLAY_PROP_PLAYSTATUS:
      {
        gint status = g_value_get_int (value);

        private->playstatus = status;
      }
      break;

    case BMP_PLAY_PROP_STREAM:
      {
        gchar *mrl;

        if (private->mrl)
          g_free (private->mrl);

        mrl = g_value_dup_string (value);

        private->mrl = mrl;
      }
      break;

    default:
      g_assert (FALSE);
      break;
    }
}

static void
bmp_play_get_property (GObject    *object,
                       guint       property_id,
                       GValue     *value,
                       GParamSpec *pspec)
{
  BmpPlay        *self    = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
    {
    case BMP_PLAY_PROP_STREAM:
      g_value_set_string (value, private->mrl);
      break;

    case BMP_PLAY_PROP_PLAYSTATUS:
      g_value_set_int (value, private->playstatus);
      break;

    default:
      g_assert (FALSE);
    }
}

static void
bmp_play_dispose (GObject * obj)
{
  BmpPlay        *self = (BmpPlay *) obj;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  if (private->dispose_has_run)
    return;

  private->dispose_has_run = TRUE;

  g_message ("stopped: %s", G_OBJECT_TYPE_NAME (obj));
}

static void
bmp_play_finalize (GObject * obj)
{
  /* nothing to do */
}

static void
bmp_play_class_init (BmpPlayClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (BmpPlayPrivate));

  gobject_class->set_property = bmp_play_set_property;
  gobject_class->get_property = bmp_play_get_property;
  gobject_class->dispose = bmp_play_dispose;
  gobject_class->finalize = bmp_play_finalize;
  gobject_class->constructor = bmp_play_constructor;

  g_object_class_install_property (gobject_class,
                                   BMP_PLAY_PROP_STREAM,
                                   g_param_spec_string ("mrl",
                                                        "mrl",
                                                        "mrl", "",
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
                                   BMP_PLAY_PROP_PLAYSTATUS,
                                   g_param_spec_int ("playstatus",
                                                     "playstatus",
                                                     "player playing status",
                                                     0, 128, 0, G_PARAM_READWRITE));

  signals[BMP_PLAY_SIGNAL_TIME_TICK] =
    g_signal_new ("time_tick",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, time_tick),
                  NULL, NULL, g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);

  signals[BMP_PLAY_SIGNAL_SET_STREAM_TITLE] =
   g_signal_new ("set_stream_title",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, set_stream_title),
                  NULL, NULL, g_cclosure_marshal_VOID__STRING,
                  G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[BMP_PLAY_SIGNAL_BUFFER] =
    g_signal_new ("buffer",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, buffer),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);

  signals[BMP_PLAY_SIGNAL_EOS] =
    g_signal_new ("eos",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, eos),
                  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                  0);

  signals[BMP_PLAY_SIGNAL_ERROR] =
    g_signal_new ("error",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, error),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);
}

BmpPlay *
bmp_play_new (void)
{
  BmpPlay *self = g_object_new (BMP_TYPE_PLAY, NULL);

  return self;
}

static GObject *
bmp_play_constructor (GType                  type,
                      guint                  n_construct_properties,
                      GObjectConstructParam *construct_properties)
{
  GObject *obj;
  BmpPlayClass *klass;
  GObjectClass *parent_class;

  klass = PLAY_CLASS (g_type_class_peek (BMP_TYPE_PLAY));
  parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));

  obj = parent_class->constructor (type,
                                   n_construct_properties,
                                   construct_properties);

  return obj;
}

static void
bmp_play_init (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  private->dispose_has_run = FALSE;
  private->mrl = NULL;
  private->stream = NULL;
  private->audioport = NULL;
  private->xine = NULL;
  private->playstatus = PLAYSTATUS_STOPPED;
  private->evq = NULL;
  private->sending = FALSE;

  playobjects_create (self);

  g_signal_connect (G_OBJECT (self), "notify::playstatus",
                    G_CALLBACK (playstatus_notify_cb), NULL);

  g_message ("running: %s", G_OBJECT_TYPE_NAME (self));
}
