/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 *
 */


#ifndef _PL_MANAGER_H_
#define _PL_MANAGER_H_

#include <config.h>
#include <glib-object.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>
#include <libhrel/relation.h>

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#endif

#if 0
#include <bmpx/list.h>
#include "bstore.h"
#endif

G_BEGIN_DECLS

#define PL_TYPE_MANAGER			  (pl_manager_get_type ())
#define PL_MANAGER(obj)			  (G_TYPE_CHECK_INSTANCE_CAST ((obj), PL_TYPE_MANAGER, PLManager))
#define PL_MANAGER_CLASS(klass)		  (G_TYPE_CHECK_CLASS_CAST ((klass), PL_TYPE_MANAGER, PLManagerClass))
#define PL_IS_MANAGER(obj)		  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), PL_TYPE_MANAGER))
#define PL_IS_MANAGER_CLASS(klass)	  (G_TYPE_CHECK_CLASS_TYPE ((klass), PL_TYPE_MANAGER))
#define PL_MANAGER_GET_CLASS(obj)	  (G_TYPE_INSTANCE_GET_CLASS ((obj), PL_TYPE_MANAGER, PLManagerClass))

typedef struct _PLManager PLManager;
typedef struct _PLManagerClass PLManagerClass;
typedef struct _PLManagerPrivate PLManagerPrivate;

struct _PLManager {
    GObject parent;

    /*
     * instance members 
     */
    PLManagerPrivate *private;
};

struct _PLManagerClass {
    GObjectClass parent;

    void    (*list_created) (PLManager *self, gchar *name);
    void    (*list_deleted) (PLManager *self, gchar *name);
};

GType
pl_manager_get_type (void);

PLManager*
pl_manager_new ();


//List creation/deletion
gboolean
pl_manager_list_new (PLManager	  *self,
		     const gchar  *name,
		     GList	  *uri_list);

gboolean
pl_manager_list_new_from_uri (PLManager	   *self,
			      const gchar  *name,
			      const gchar  *uri);
 
gboolean
pl_manager_list_delete (PLManager	  *self,
			const gchar	  *name);


//List manipulation
gboolean
pl_manager_list_remove_uris (PLManager	  *self,
			     const gchar  *name,
			     GList	  *uri_list);

//List accessors
const HRelation*
pl_manager_list_get_relation (PLManager	      *self,
			      const gchar     *name);


//Meta
GList*
pl_manager_get_lists (PLManager *self);

#ifdef HAVE_GUI
GtkListStore*
pl_manager_list_get_model (PLManager	*self,
			   const gchar  *name);

#endif

G_END_DECLS

#endif /* _PL_MANAGER_H_ */
