/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

/*
 * This file contains code originally taken from XMMS-scrobbler:
 *  
 * XMMS-scrobbler (C) GPL by
 *
 * Audun Hove <audun@nlc.no>
 * Pipian <pipian@pipian.com>
 *
 */

/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include <bmpx/util.h>
#include <bmpx/uri.h>
#include <bmpx/vfs.h>

#include <goa/libgoa.h>

#include "system_control.h"
#include "main.h"

#include "scrobbler.h"
#include "md5.h"

#include <glib/gi18n.h>

#define BMPX_SCROBBLER_URI_SCHEME   "http"
#define BMPX_SCROBBLER_URI_HOST     "post.audioscrobbler.com"
#define BMPX_SCROBBLER_URI_PORT	     80
#define BMPX_SCROBBLER_USERAGENT    "BMPx/0.13"

#define BMPX_SCROBBLER_URI_PATH     "/?hs=true&p=1.1&c=%s&v=%s&u=%s"

#define RESP_UPTODATE		    "UPTODATE"
#define RESP_UPDATE		    "UPDATE"
#define RESP_FAILED		    "FAILED"
#define RESP_BADUSER		    "BADUSER"

#define BMPX_SCROBBLER_CLIENT_ID "mpx"
#define BMPX_SCROBBLER_CLIENT_VERSION "1.0"

typedef enum {
    SCROBBLER_HANDSHAKE_UPTODATE,
    SCROBBLER_HANDSHAKE_UPDATE,
    SCROBBLER_HANDSHAKE_BADUSER,
    SCROBBLER_HANDSHAKE_FAILED,

    SCROBBLER_HANDSHAKE_NETWORK_ERROR,
    SCROBBLER_HANDSHAKE_STATE_UNDEFINED

} ScrobblerHandshakeStatus;

typedef struct _ScrobblerHandshakeInfo ScrobblerHandshakeInfo;
struct _ScrobblerHandshakeInfo {
	gboolean     uptodate;
	gchar	    *uri;
	gchar	    *md5;
	gint	     interval;
	gchar	    *info;	// Used for FAILED <reason>, and UPDATE url field
};

G_DEFINE_TYPE(BmpScrobbler, bmp_scrobbler, G_TYPE_OBJECT)

#if 0
typedef enum {
	BMP_SCROBBLER_N_SIGNALS
} BmpScrobblerSignals;
static guint bmp_scrobbler_signals[BMP_SCROBBLER_N_SIGNALS] = {};
#endif

typedef enum {
	BMP_SCROBBLER_N_PROPERTIES

} BmpScrobblerProperties;

enum {
	H_TRACK_CHANGE,	
	H_SET_STREAM_POS,
	H_LASTFM_ENABLE,

	N_HANDLERS
};

struct _BmpScrobblerPrivate
{
        gboolean dispose_has_run;

	ScrobblerHandshakeInfo *info;

	gboolean		enabled;
	gboolean		sent;

	gint			duration;
	gint			position_previous;

	gchar		       *md5_response;

	gulong			handler[N_HANDLERS];

	gint			message_domain_id;
};

static void
scrobbler_run_handshake (BmpScrobbler *self);

static char*
get_timestr (time_t t, int gmt);
	
static void
scrobbler_track_change (BmpSystemControl    *system_control,
			BmpScrobbler	    *self)
{
    //Determine the length of the current song
    gchar	*current_uri;
    GError	*error;
    GHashTable	*metadata;
    GValue  	*value;

    error = NULL;
    metadata = NULL;

    if (!bmp_system_control_get_current_uri (system_control, &current_uri, &error)) return;
    if (!bmp_system_control_get_metadata_for_uri (system_control, current_uri, &metadata, &error)) return; 

    value = g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_TIME));
    self->private->duration = g_value_get_int (value);
    g_hash_table_destroy (metadata);

    self->private->sent = FALSE;
    self->private->position_previous = 0; 
}

/* Stolen from libtransport_http */
typedef struct _SCReadHandle SCReadHandle;
struct _SCReadHandle {
    GMutex  *handle_lock;
    char    *memory;
    size_t   size;
};
#define SC_READ_HANDLE(_data) ((SCReadHandle*)(_data))

static int
read_block (void *data, const char *buf, size_t len)
{
  SCReadHandle *rhandle = SC_READ_HANDLE(data);

  if (!len)
    {
        g_mutex_unlock (rhandle->handle_lock);
        return 0;
    }

  rhandle->memory = realloc (rhandle->memory, rhandle->size+len);
  memcpy  (rhandle->memory+rhandle->size, buf, len);
  rhandle->size += len;

  return 0;
}

static void
scrobbler_send_song_information (BmpScrobbler	    *self,
				 gint		     position)
{
#define SCROBBLER_SONG_POST_STRING "u=%s&s=%s&a[0]=%s&t[0]=%s&m[0]=&b[0]=%s&l[0]=%d&i[0]=%s"
    
    SCReadHandle    *rhandle;
    gchar	    *current_uri;
    GError	    *error;
    GHashTable	    *metadata;
    GValue	    *value;
    gchar	    *post_data;
    gchar	    *timestr;
    gchar	   **resp, **line;
    const gchar	    *artist,
		    *title,
		    *album;
    gint	     duration;
    ne_session	    *sess;
    ne_request	    *reqs;
    GURI	    *uri;
    gchar	    *user;

    self->private->sent = TRUE;

    error = NULL;
    metadata = NULL;

    if (!bmp_system_control_get_current_uri (bmp_system_control, &current_uri, &error)) return;
    if (!bmp_system_control_get_metadata_for_uri (bmp_system_control, current_uri, &metadata, &error)) return; 

    if (!g_uri_test (current_uri, G_URI_TEST_PROTOCOL, PROTOCOL_FILE))
	{
	    g_free (current_uri);
	    return;
	}

    value = g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_TIME));
    duration = g_value_get_int (value);

    if (duration < 30)
	{
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Not submitting '%s' to scrobbler: duration of less than 30 seconds",
						    G_STRLOC,
						    current_uri);
	    g_free (current_uri);
	    return;
	}

    value = g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_ARTIST));
    artist = g_value_get_string (value);

    value = g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_ALBUM));
    album = g_value_get_string (value);

    value = g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_TITLE));
    title = g_value_get_string (value);

    if (!duration || !artist || !title)
	{
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Not submitting '%s' to scrobbler: missing metadata!",
						    G_STRLOC,
						    current_uri);
	    g_free (current_uri);
	    return;
	}

    timestr = get_timestr (time(NULL),1);
    user = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "username");

    post_data = g_strdup_printf (SCROBBLER_SONG_POST_STRING,
				   user, 
				   self->private->md5_response,
				   ne_path_escape(artist),
				   ne_path_escape(title),
				   ne_path_escape(album),
				   duration,
				   timestr);

    g_message ("%s: Submitting song '%s' with POST data '%s'", G_STRLOC, current_uri, post_data);

    if (!(uri = bmp_uri_new (self->private->info->uri)))
	return;
	
    sess = ne_session_create (uri->scheme,
			      uri->hostname, 
			      80);

    ne_set_useragent (sess, "BMPx/0.13");
    ne_set_read_timeout (sess, 30); 
    reqs = ne_request_create (sess, "POST", uri->path);
    ne_add_request_header (reqs, "Content-Type", "application/x-www-form-urlencoded");

    ne_set_request_body_buffer (reqs, (const char *)post_data, (size_t)strlen(post_data));

    rhandle = g_new0 (SCReadHandle, 1);
    rhandle->memory = NULL;
    rhandle->size = 0;
    rhandle->handle_lock = g_mutex_new ();

    ne_add_response_body_reader (reqs, ne_accept_2xx, read_block, rhandle);

    g_mutex_lock (rhandle->handle_lock);

    if (ne_request_dispatch (reqs) != NE_OK)
	{
	    gchar *str;

	    g_mutex_unlock (rhandle->handle_lock);
	    g_free (rhandle);
	    ne_request_destroy (reqs);
	    ne_session_destroy (sess);

	    str = g_strdup_printf ("%s: '%s'", _("A Network error occured while trying to connect to Last.FM"), ne_get_error (sess));

	    bmp_system_control_message_dispatch (bmp_system_control,
					     self->private->message_domain_id, 
					     GTK_MESSAGE_ERROR,
					     str);

	    xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
	    g_free (str);
	}

    while (!g_mutex_trylock(rhandle->handle_lock))
      {
        while (gtk_events_pending ())
          gtk_main_iteration ();
      }

    g_mutex_unlock (rhandle->handle_lock);

    ne_request_destroy (reqs);
    ne_session_destroy (sess);

    rhandle->memory = realloc (rhandle->memory, rhandle->size + 1);
    rhandle->memory[rhandle->size+1] = 0;

    resp = g_strsplit (rhandle->memory, "\n", -1);

    if (!strcmp("BADAUTH",resp[0]))
	{
	    bmp_system_control_message_dispatch (bmp_system_control,
					     self->private->message_domain_id, 
					     GTK_MESSAGE_ERROR,
					    _("There was an authentication error during song submission.\n\nPlease check your Last.FM username/password.\n\nLast.FM support has to be disabled until proper credentials are provided.")); 
	    xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
	}
    else if (!strncmp("FAILED ", resp[0], 7))
	{
	    g_message("%s: Submitting song '%s' failed: %s", G_STRLOC, current_uri, resp[0]);
	}
    else /* OK assumed */
	{
	    g_message ("%s: Submitting song '%s' succeeded.", G_STRLOC, current_uri);
	}

    g_strfreev(resp);

#if 0
    g_free (timestr);
    g_free (hs_result);
    g_free (rhandle);
    //Destroy mutex 
    g_hash_table_destroy (metadata);
#endif
}


static void
scrobbler_set_stream_pos	(BmpSystemControl   *system_control,				
				 gint		     position,
				 BmpScrobbler	    *self) 
{
    gboolean enabled = xcs_key_fetch_boolean (xcs, "lastfm", "enable"); 
    gboolean playing = (g_object_get_int (bmp_play_engine, "playstatus") == PLAYSTATUS_PLAYING) ? TRUE : FALSE;

    if ((position - self->private->position_previous) > 4) // Well, a rather lame check for a seek
	{
	    self->private->sent = TRUE;
	    return;
	}

    self->private->position_previous = position;

    if (enabled && playing && (!self->private->sent))
	{
	    if ((position >= 240) || (position >= (self->private->duration/2)))
		{
		    scrobbler_send_song_information (self, position);
		}
	}
}

static void
scrobbler_lastfm_enable_changed (Xcs		    *config,
				 BmpScrobbler	    *self)
{
    gboolean enabled = xcs_key_fetch_boolean (xcs, "lastfm", "enable");
    if (enabled) scrobbler_run_handshake (self);
}

/* GObject */
static void
bmp_scrobbler_init (BmpScrobbler *self)
{
    self->private = g_new (BmpScrobblerPrivate, 1);
    self->private->dispose_has_run = FALSE;

    self->private->info = NULL;
    self->private->sent = FALSE;
    self->private->duration = 0;
    self->private->md5_response = NULL;

}

static void
bmp_scrobbler_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id)
	    {
		default: break;
	    }
}

static void
bmp_scrobbler_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id)
	    {
		default: break;
	    }
}

static GObject *
bmp_scrobbler_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpScrobblerClass *klass;
                GObjectClass *parent_class;
                klass = BMP_SCROBBLER_CLASS (g_type_class_peek (BMP_TYPE_SCROBBLER));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }


        return obj;
}

static void
bmp_scrobbler_dispose (GObject *obj)
{
        BmpScrobbler *self = (BmpScrobbler*) obj;

        if (self->private->dispose_has_run) return;
        self->private->dispose_has_run = TRUE;

}

static void
bmp_scrobbler_finalize (GObject *obj)
{
        BmpScrobbler *self = (BmpScrobbler*)obj;

        g_free (self->private);
}

BmpScrobbler*
bmp_scrobbler_new (BmpSystemControl *system_control)
{
  BmpScrobbler *self;

  self = g_object_new (bmp_scrobbler_get_type (), NULL);

  self->private->handler[H_TRACK_CHANGE] =
			g_signal_connect (G_OBJECT(system_control),
				          "track-change",
					  G_CALLBACK(scrobbler_track_change),
					  self);

  self->private->handler[H_SET_STREAM_POS] =
			g_signal_connect (G_OBJECT(system_control),
				          "set-stream-pos",
					  G_CALLBACK(scrobbler_set_stream_pos),
					  self);

  self->private->handler[H_LASTFM_ENABLE] =
			g_signal_connect (G_OBJECT(xcs),	
					  "changed::lastfm-enable",
					  G_CALLBACK(scrobbler_lastfm_enable_changed),
					  self);

  self->private->message_domain_id =
			 bmp_system_control_message_domain_register (system_control,
								     "lastfm", 
								     _("BMPx Last.FM"));

  return self;
}

static void
bmp_scrobbler_class_init (BmpScrobblerClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = bmp_scrobbler_set_property;
        gobject_class->get_property = bmp_scrobbler_get_property;
        gobject_class->dispose	    = bmp_scrobbler_dispose;
        gobject_class->finalize	    = bmp_scrobbler_finalize;
        gobject_class->constructor  = bmp_scrobbler_constructor;
}

/* Scrobbler */
static char*
hexify (char *pass, int len)
{
    char *hash;
    char *bp;
    char hexchars[] = "0123456789abcdef";
    int i;

    hash = g_malloc0 (33);
    bp = hash;

    for(i = 0; i < len; i++)
	{
	    *(bp++) = hexchars[(pass[i] >> 4) & 0x0f];
	    *(bp++) = hexchars[pass[i] & 0x0f];
	}
    *bp = 0;

    return hash;
}

static char*
get_timestr (time_t t, int gmt)
{
	struct tm *tm;
	static char buf[30];

	tm = gmt ? gmtime(&t) : localtime(&t);
	snprintf(buf, sizeof(buf), "%d-%.2d-%.2d %.2d:%.2d:%.2d",
			tm->tm_year + 1900,
			tm->tm_mon + 1,
			tm->tm_mday,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec);
	return buf;
}

static ScrobblerHandshakeStatus
scrobbler_handshake (BmpScrobbler *self)

{
    SCReadHandle    *rhandle;
    ne_session	    *sess;
    ne_request	    *reqs;
    gchar	    *path,
		    *hs_result;
    gchar	   **resp, **line;
    gint	     hs_length;
    gchar	    *user;

    self->private->info = g_new0 (ScrobblerHandshakeInfo,1);

    user = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "username");

    path = g_strdup_printf (BMPX_SCROBBLER_URI_PATH,
			    BMPX_SCROBBLER_CLIENT_ID,
			    BMPX_SCROBBLER_CLIENT_VERSION,
			    user); 

    // g_free (user);

    sess = ne_session_create (BMPX_SCROBBLER_URI_SCHEME,
			      BMPX_SCROBBLER_URI_HOST,
			      BMPX_SCROBBLER_URI_PORT);

    ne_set_useragent (sess, "BMPx/0.13");
    ne_set_read_timeout (sess, 30); 
    reqs = ne_request_create (sess, "GET", path);

    rhandle = g_new0 (SCReadHandle, 1);
    rhandle->memory = NULL;
    rhandle->size = 0;
    rhandle->handle_lock = g_mutex_new ();
    g_mutex_lock (rhandle->handle_lock);

    ne_add_response_body_reader (reqs, ne_accept_2xx, read_block, rhandle);

    if (ne_request_dispatch (reqs) != NE_OK)
	{
	    gchar *str;

	    g_mutex_unlock (rhandle->handle_lock);
	    g_free (rhandle);
	    ne_request_destroy (reqs);
	    ne_session_destroy (sess);

	    str = g_strdup_printf ("%s: '%s'", _("A Network error occured while trying to connect to Last.FM"), ne_get_error (sess));

	    bmp_system_control_message_dispatch (bmp_system_control,
					     self->private->message_domain_id, 
					     GTK_MESSAGE_ERROR,
					     str);

	    xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
	    g_free (str);

	    return SCROBBLER_HANDSHAKE_NETWORK_ERROR;
	}

    while (!g_mutex_trylock(rhandle->handle_lock))
      {
        while (gtk_events_pending ())
          gtk_main_iteration ();
      }

    g_mutex_unlock (rhandle->handle_lock);

    ne_request_destroy (reqs);
    ne_session_destroy (sess);

    resp = g_strsplit (rhandle->memory, "\n", -1);
    // g_free (rhandle->memory);

    //Parse response 
    if (!resp)
	{
	    return SCROBBLER_HANDSHAKE_NETWORK_ERROR; //Not 100% true, but also not entirely wrong
	}

    line = g_strsplit (resp[0], " ", -1);

    if (!strcmp(line[0], RESP_UPTODATE)) 
	{
	    self->private->info->uptodate = TRUE;
	    g_strfreev (line);

	    self->private->info->md5 = g_strdup (resp[1]);	    
	    self->private->info->uri = g_strdup (resp[2]);

	    line = g_strsplit (resp[3], " ", -1);

	    if (line[0] && line[1])
		self->private->info->interval = atoi (line[1]);

	    g_strfreev (line);

	    return SCROBBLER_HANDSHAKE_UPTODATE;
    
	}
    else
    if (!strcmp(line[0], RESP_UPDATE))
	{
	    self->private->info->uptodate = FALSE;
	    self->private->info->info = g_strdup (line[1]);

	    self->private->info->md5 = g_strdup (resp[1]);	    
	    self->private->info->uri = g_strdup (resp[2]);

	    line = g_strsplit (resp[3], " ", -1);

	    if (line[0] && line[1])
		self->private->info->interval = atoi (line[1]);

	    g_strfreev (line);

	    return SCROBBLER_HANDSHAKE_UPDATE;
	}
    else 
    if (!strcmp(line[0], RESP_BADUSER))
	{
	    g_strfreev (line);

	    line = g_strsplit (resp[1], " ", -1);

	    if (line[0] && line[1])
		self->private->info->interval = atoi (line[1]);

	    g_strfreev (line);
	    g_strfreev (resp);

	    return SCROBBLER_HANDSHAKE_BADUSER;
	}
    if (!strcmp(line[0], RESP_FAILED))
	{
	    self->private->info->info = g_strdup (line[1]);
	    g_strfreev (line);

	    line = g_strsplit (resp[1], " ", -1);

	    if (line[0] && line[1])
		self->private->info->interval = atoi (line[1]);

	    g_strfreev (line);
	    g_strfreev (resp);

	    return SCROBBLER_HANDSHAKE_FAILED;
	}
	
    g_strfreev (line);
    g_strfreev (resp);
    g_free (rhandle);

    return SCROBBLER_HANDSHAKE_STATE_UNDEFINED;
}

static void
scrobbler_run_handshake (BmpScrobbler *self)
{
    ScrobblerHandshakeStatus hs_result;
    gchar *user, *pass;

    user = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "username");
    pass = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "password");

    if (!user || !pass)
	{
	    //FIXME: Error here

	    return;
	}

    hs_result = scrobbler_handshake (self);

    switch (hs_result)
	{
		case SCROBBLER_HANDSHAKE_UPTODATE:
		    {
			md5_state_t  md5state;
			char         md5pword[16];
			char	    *str;
			char	    *pass, *aux;

			g_message ("%s: UPTODATE. MD5:'%s', URL:'%s', INTERVAL:'%d'", G_STRLOC,
										      self->private->info->md5,
										      self->private->info->uri,
										      self->private->info->interval);

			pass = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "password");

		        md5_init (&md5state);
			md5_append (&md5state, (unsigned const char *)pass, strlen(pass));
			md5_finish (&md5state, (unsigned char *)md5pword);
			aux = hexify (md5pword, sizeof(md5pword));

			str = g_strconcat (aux, self->private->info->md5, NULL);
		        md5_init (&md5state);
			md5_append (&md5state, (unsigned const char *)str, strlen(str));
			md5_finish (&md5state, (unsigned char *)md5pword);

			self->private->md5_response = hexify (md5pword, sizeof(md5pword));

			g_free (aux);
			g_free (str);
		    }
		    break;

		case SCROBBLER_HANDSHAKE_UPDATE:
		    {
			md5_state_t  md5state;
			char         md5pword[16];
			char	    *str;
			char	    *pass, *aux;

			g_message ("%s: UPDATE. UPDATE_URL:'%s', MD5:'%s', URL:'%s', INTERVAL:'%d'", G_STRLOC,
										      self->private->info->info,
										      self->private->info->md5,
										      self->private->info->uri,
										      self->private->info->interval);

			pass = (gchar*)xcs_key_fetch_string (xcs, "lastfm", "password");

		        md5_init (&md5state);
			md5_append (&md5state, (unsigned const char *)pass, strlen(pass));
			md5_finish (&md5state, (unsigned char *)md5pword);
			aux = hexify (md5pword, sizeof(md5pword));

			str = g_strconcat (aux, self->private->info->md5, NULL);
		        md5_init (&md5state);
			md5_append (&md5state, (unsigned const char *)str, strlen(str));
			md5_finish (&md5state, (unsigned char *)md5pword);

			self->private->md5_response = hexify (md5pword, sizeof(md5pword));

			g_free (aux);
			g_free (str);

		    }
		    break;

		case SCROBBLER_HANDSHAKE_BADUSER:
		    {
			g_message ("%s: BADUSER. INTERVAL:'%d'", G_STRLOC, 
						  self->private->info->interval);

			bmp_system_control_message_dispatch	    (bmp_system_control,
						    		     self->private->message_domain_id, 
								     GTK_MESSAGE_ERROR,
								     _("Please check your Last.FM username/password!\n\nLast.FM support has to be disabled until proper credentials are provided.")); 
			xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
		    }
		    break;

		case SCROBBLER_HANDSHAKE_FAILED:
		    {
			gchar *aux;

			g_message ("%s: FAILED. REASON:'%s', INTERVAL:'%d'", G_STRLOC, 
						  self->private->info->info,
						  self->private->info->interval);

			aux = g_strdup_printf (_("Connection with Last.FM failed. Reason given is: '%s'"),
						self->private->info->info);

			bmp_system_control_message_dispatch	    (bmp_system_control,
						    		     self->private->message_domain_id, 
								     GTK_MESSAGE_ERROR,
								     aux); 

			g_free (aux);
			xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
		    }
		    break;

		case SCROBBLER_HANDSHAKE_NETWORK_ERROR:
		    {
			g_message ("%s: NETWORK ERROR", G_STRLOC); 

			bmp_system_control_message_dispatch	    (bmp_system_control,
						    		     self->private->message_domain_id, 
								     GTK_MESSAGE_ERROR,
								     _("A network error occured during connection handshake with Last.FM")); 


			xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
		    }
		    break;

		case SCROBBLER_HANDSHAKE_STATE_UNDEFINED:
		    {
			g_message ("%s: STATE UNDEFINED", G_STRLOC); 

			//FIXME: Error message
			xcs_key_set_boolean (xcs, "lastfm", "enable", FALSE);
		    }
		    break;

		default:
			g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unknown Scrobbler handshake status response!", G_STRLOC);
			break;
	}
}
