/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2003  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 2000-2003  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "main.h"

#include "ui_util.h"
#include "ui_dialog_about.h"
#include "wm.h"

#include "signal_handler_pool.h"

static GtkWidget	  *about_window = NULL;
static GdkPixbuf	  *about_pixbuf = NULL;
static SignalHandlerPool  *pool = NULL;

static gboolean
on_about_window_key_press (GtkWidget *widget, GdkEventKey *event, gpointer data)
{
  g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

  if (event->keyval == GDK_Escape)
    {      
      gtk_widget_hide (widget);
      wm_about_stop ();
    }

  return FALSE;
}

static gboolean
on_about_window_expose (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  cairo_t *cr;

  cr = gdk_cairo_create (widget->window);
  
  gdk_cairo_set_source_pixbuf (cr, about_pixbuf, 0, 0);

  cairo_reset_clip (cr);
  gdk_cairo_region (cr, event->region);
  cairo_clip (cr);
  cairo_paint (cr);

  cairo_destroy (cr);

  return FALSE;
}

void
dialog_about_destroy (void)
{
  if (about_window)
    {
      g_object_unref (about_pixbuf);
      signal_handler_pool_free (pool);
      gtk_widget_destroy (about_window);
      about_window = NULL;
    }
}
 
void
dialog_about_setup (void)
{
  GdkPixmap *mask_pixmap_window1 = NULL,
	    *mask_pixmap_window2 = NULL;
  GdkBitmap *mask_bitmap_window1 = NULL,
	    *mask_bitmap_window2 = NULL;
  gchar	    *filename;

  about_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  RM_ITEM_NEW("windows", "window-dialog-about", about_window, NULL);

  gtk_window_set_type_hint(GTK_WINDOW(about_window), GDK_WINDOW_TYPE_HINT_DIALOG);

  filename = DATA_DIR G_DIR_SEPARATOR_S "images" G_DIR_SEPARATOR_S "about-logo.png";
  about_pixbuf = gdk_pixbuf_new_from_file (filename, NULL);

  gtk_widget_set_size_request (GTK_WIDGET (about_window),
			       gdk_pixbuf_get_width (about_pixbuf),
			       gdk_pixbuf_get_height (about_pixbuf));

  gdk_pixbuf_render_pixmap_and_mask (about_pixbuf,
				     &mask_pixmap_window1,
				     &mask_bitmap_window1,
				     0);

  gdk_pixbuf_render_pixmap_and_mask (about_pixbuf,
				     &mask_pixmap_window2,
				     &mask_bitmap_window2,
				     128);

  gtk_widget_add_events (about_window, GDK_ALL_EVENTS_MASK);

  pool = signal_handler_pool_new ("UiAbout");

  signal_handler_pool_add_connection (pool,
				      about_window,	
				      "key-press-event",	
				      G_CALLBACK (on_about_window_key_press),	
				      NULL,
				      0);	

  signal_handler_pool_add_connection (pool,
				      about_window,	
				      "expose-event",	
				      G_CALLBACK (on_about_window_expose),	
				      NULL,
				      0);	

  signal_handler_pool_add_connection (pool,
				      about_window,	
				      "delete-event",	
				      G_CALLBACK (gtk_widget_hide),	
				      about_window,
				      G_CONNECT_SWAPPED);	

  gtk_widget_set_app_paintable (about_window, TRUE);
  gtk_widget_shape_combine_mask (GTK_WIDGET (about_window), mask_bitmap_window2, 0, 0);
  gtk_window_set_decorated (GTK_WINDOW (about_window), FALSE);
  gtk_window_set_title (GTK_WINDOW (about_window), "About BMPx");
  gtk_window_set_position (GTK_WINDOW (about_window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (about_window), FALSE);
  bmp_window_set_icon_list (GTK_WIDGET(about_window), "player");

  g_object_unref (mask_bitmap_window1); 
  g_object_unref (mask_bitmap_window2); 
  g_object_unref (mask_pixmap_window1); 
  g_object_unref (mask_pixmap_window2); 
}

void
bmp_about_dialog_show (void)
{
  wm_about_start ();
  gtk_window_present (GTK_WINDOW(about_window));
}

