/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 *
 */


#include <config.h>
#include <revision.h>
#include <build.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include <bmpx/util.h>

#include <libskinned/bmp_tooltips.h>
#include <libchroma/chroma-list.h>

#include <xcs/xcs.h>

#include "glade.h"
#include "ui.h"
#include "ui_main.h"
#include "ui_playlist.h"
#include "ui_util.h"
#include "loader.h"

#include "ui_dialog_preferences.h"
#include "signal_handler_pool.h"
#include "plugin_host.h"


#include "main.h"

#define THUMBNAIL_WIDTH  90
#define THUMBNAIL_HEIGHT 40

enum SkinViewCols {
    SKIN_VIEW_COL_PREVIEW,
    SKIN_VIEW_COL_NAME,
    SKIN_VIEW_COL_PATH,
    SKIN_VIEW_N_COLS
};

enum CategoryViewCols {
    CATEGORY_VIEW_COL_ICON,
    CATEGORY_VIEW_COL_NAME,
    CATEGORY_VIEW_COL_ID,

    CATEGORY_VIEW_N_COLS
};

#define SKIN_NODE(x)  ((SkinNode *)(x))
struct _SkinNode {
    gchar *name;
    gchar *path;
    GTime *time;
};
typedef struct _SkinNode SkinNode;

typedef struct {
    const gchar *icon_path;
    const gchar *name;
    gint id;
} Category;

typedef struct {
    const gchar *name;
    const gchar *tag;
}
TitleFieldTag;

static GHashTable *map_skins;

static Category categories[] = {
    {DATA_DIR "/images/preferences/appearance.png", N_("Appearance"),   0},
/*    {DATA_DIR "/images/preferences/eq.png",         N_("Equalizer"),	1}, */
    {DATA_DIR "/images/preferences/playlist.png",   N_("Playlist"),	2},
    {DATA_DIR "/images/preferences/playback.png",   N_("Playback"),	3}, 
    {DATA_DIR "/images/preferences/plugins.png",    N_("Plugins"),	4},
    {DATA_DIR "/images/preferences/lastfm.png",     N_("Last.FM"),	5}
};
static gint n_categories = G_N_ELEMENTS(categories);

static BmpTooltipEntry window_preferences_tooltips[] = {
    { "titlestring_entry",
      N_("Allows you to customize how text is displayed in the playlist"),
      GTK_STOCK_INFO },

    { "skin_view",
      N_("Here you can choose a skin to determine how BMPx looks like"),
      GTK_STOCK_INFO },

    { "font_per_skin",
      N_("Enabling this will make BMPx remember the font for each skin individually, giving you more ease when theming BMPx"),
      GTK_STOCK_INFO },

    { "font_button",
      N_("Click on this button to select a font for BMPx to use"),
      GTK_STOCK_INFO }
};
static gint n_window_preferences_tooltips = G_N_ELEMENTS (window_preferences_tooltips);

#if 0
static gchar* eq_sliders[] = {
    "eq_30",
    "eq_60",
    "eq_125",
    "eq_250",
    "eq_500",
    "eq_1000",
    "eq_2000",
    "eq_4000",
    "eq_8000",
    "eq_16000"
};
static gint n_eq_sliders = G_N_ELEMENTS(eq_sliders);
#endif

static void
update_skinlist(BmpPreferences *self);

static GdkPixbuf *
skin_get_thumbnail(const gchar * path);

/* GObject stuff */
G_DEFINE_TYPE (BmpPreferences, bmp_preferences, G_TYPE_OBJECT)

struct _BmpPreferencesPrivate {

    gboolean		   dispose_has_run;

    GladeXML		  *xml;
    gchar		  *skins_path;
    GList		  *skinlist;

    GtkWidget		  *skin_view;
    GtkWidget		  *skin_view_notebook;
  
    GtkWidget		  *columns_view;
    GtkListStore	  *columns_store;

    BmpPluginHost	  *host;
    SignalHandlerPool	  *pool;
    gboolean		   updated;
};

void
window_preferences_create (BmpPreferences *self);

static void
set_font (BmpPreferences *self, const gchar *font_name)
{
    GtkWidget *font_button;

    font_button = glade_xml_get_widget (self->private->xml, "font_button");
    gtk_font_button_set_font_name(GTK_FONT_BUTTON(font_button), font_name);
}

BmpPreferences*
bmp_preferences_new (void)
{
    BmpPreferences *self;

    self = g_object_new (BMP_TYPE_PREFERENCES, NULL);
    return self;
}

static GObject *
bmp_preferences_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /*
         * Invoke parent constructor.
         */
        BmpPreferencesClass *klass;
        GObjectClass *parent_class;
        klass = BMP_PREFERENCES_CLASS (g_type_class_peek (BMP_TYPE_PREFERENCES));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    /*
     * do stuff.
     */
    return obj;
}

static void
bmp_preferences_dispose (GObject *object)
{
    BmpPreferences *self = (BmpPreferences *)object;

    signal_handler_pool_free (self->private->pool);
    if (self->private->xml) g_object_unref (self->private->xml);
    if (self->window) gtk_widget_destroy (self->window);
}

static void
bmp_preferences_class_init (BmpPreferencesClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->dispose     = bmp_preferences_dispose;
    gobject_class->constructor = bmp_preferences_constructor;

}

static void
bmp_preferences_init (BmpPreferences *self)
{
    self->private = g_new (BmpPreferencesPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->skins_path = NULL;
    self->private->skinlist = NULL;
    self->private->host = bmp_plugin_host_new ();
    self->private->updated = FALSE;
    self->private->pool = signal_handler_pool_new ("UiPreferences");

    window_preferences_create (self);
}

/* Our stuff */
static void
change_category(GtkNotebook * notebook,
                GtkTreeSelection * selection)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gint index;

    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, CATEGORY_VIEW_COL_ID, &index, -1);
    gtk_notebook_set_current_page(notebook, index);
}

static void
titlestring_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, gtk_combo_box_get_active_text (GTK_COMBO_BOX(object)));
}

static void
fontbutton_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, gtk_font_button_get_font_name (GTK_FONT_BUTTON(object)));
}

static void
gtkentry_binding (GObject *object, GValue *value)
{
  const gchar *text;

  text = gtk_entry_get_text (GTK_ENTRY(object));
    
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, text); 
}

static void
on_font_button_font_set(GtkFontButton * button,
                                 gpointer data)
{
    const gchar *fontname;

    fontname = gtk_font_button_get_font_name (button);

    /* Now the font-per-skin handling */
    if (xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin"))
    {
	    gchar *basename,
		  *basename_archive;

#ifndef USE_GCONF
	    xcs_domain_new (xcs, "skin-fonts");
#endif

	    basename = 
	      g_path_get_basename (xcs_key_fetch (xcs, "bmp", "skin"));
	    basename_archive = 
	      archive_basename (basename);

	    xcs_key_set_string (xcs, "skin-fonts", basename_archive, fontname);
    }
    else
    {
	    xcs_key_set_string (xcs, "bmp", "font-save", fontname);
    }
}

void
on_font_button_realize(GtkFontButton * button,
                                gpointer data)
{
    gtk_font_button_set_font_name(button,  xcs_key_fetch (xcs, "bmp", "font"));
}

static void
on_custom_cursors_realize(GtkToggleButton *button,
                          gpointer data)
{
    gtk_toggle_button_set_active (button, xcs_key_fetch_boolean (xcs, "bmp","use-custom-cursors"));
}

static void
on_custom_cursors_toggled(GtkToggleButton *button,
                          gpointer data)
{
    BmpWindowMain *bmp_main;
    BmpWindowPlaylist *playlist;

    bmp_main = RM_GET_ITEM("windows","window-main", NOLOCK);
    playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);

    bmp_window_main_configure (bmp_main);
    bmp_window_playlist_configure (playlist);
}

static void
on_display_tooltips_realize (GtkToggleButton *button,
			     gpointer	      data)
{
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips");
    gtk_toggle_button_set_active (button, value);
}

static void
font_changed (GtkFontButton	   *button,
              gpointer		    data)
{
    const gchar* value;

    value = xcs_key_fetch (XCS_CONFIGURATION(data), "bmp", "font");

    gtk_font_button_set_font_name(GTK_FONT_BUTTON(button), value);
}

static void
use_custom_cursors_changed (GtkToggleButton      *button,
                            gpointer	          data)
{
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "use-custom-cursors");
    gtk_toggle_button_set_active (button, value);
}

static void
font_per_skin_changed (GtkToggleButton      *button,
                       gpointer	             data)
{
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin");
    gtk_toggle_button_set_active (button, value);
}

static void
skin_changed (BmpPreferences  *self,
              gpointer	             data)
{
    GtkTreeModel      *model;
    GtkTreeSelection  *selection;
    GtkTreeIter	       iter;
    GtkTreePath	      *path;
    const gchar	      *value;

    model = gtk_tree_view_get_model (GTK_TREE_VIEW(self->private->skin_view));
    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(self->private->skin_view));
    value = xcs_key_fetch (xcs, "bmp", "skin");
    attr_row_map_mapping_get (map_skins, value, model, &iter);

    gtk_tree_selection_select_iter (selection, &iter); 

    path = gtk_tree_model_get_path (model, &iter);

    gtk_tree_view_scroll_to_cell    (GTK_TREE_VIEW(self->private->skin_view),
                                     path, 
                                     NULL,
				     TRUE,
				     0.5, 0.0); 
    gtk_tree_path_free (path);
}


static void
display_tooltips_changed (GtkToggleButton      *button,
                          gpointer	        data)
{
    BmpTooltips *tooltips;
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips");

    tooltips = RM_GET_ITEM ("tooltips","tooltips-preferences", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-preferences");

    tooltips = RM_GET_ITEM ("tooltips","tooltips-main", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-main");

    gtk_toggle_button_set_active (button, value);
}

static void
on_display_numbers_realize (GtkToggleButton *button,
                            gpointer	     data)
{
    gtk_toggle_button_set_active (button,  xcs_key_fetch_boolean (xcs, "bmp", "display-tracklist-numbers"));
}

static void
display_tracklist_numbers_changed (BmpPreferences *button,
				   gpointer	   data)
{
    chroma_list_set_use_numbers (CHROMA_LIST(c_tracklist), xcs_key_fetch_boolean (xcs, "bmp", "display-tracklist-numbers"));
}

static void
on_font_per_skin_realize (GtkToggleButton *button,
                          gpointer	   data)
{
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin");
    gtk_toggle_button_set_active (button, value);
}

static void
on_font_per_skin_toggled (GtkToggleButton      *button,
                          BmpPreferences *self)
{
    const gchar *skin_font;
    gchar *basename, *basename_archive;
    gboolean value;

    value =
      gtk_toggle_button_get_active (button);

    if (value == TRUE)
    {
	const gchar *aux;

	aux = xcs_key_fetch (xcs, "bmp", "font-save");
	if (!aux)
	{
	    aux = xcs_key_fetch (xcs, "bmp", "font");
	    xcs_key_set_string (xcs, "bmp", "font-save", aux);
	}

#ifndef USE_GCONF
	xcs_domain_new (xcs, "skin-fonts");
#endif

	basename = 
	  g_path_get_basename (xcs_key_fetch (xcs, "bmp", "skin"));
	basename_archive = 
	  archive_basename (basename);

	skin_font = xcs_key_fetch (xcs, "skin-fonts", basename_archive);

	if (skin_font)
	{
	    set_font (self, skin_font);
	    xcs_key_set_string (xcs, "bmp", "font", skin_font);
	}
	else
	{
	    skin_font = xcs_key_fetch (xcs, "bmp", "font-save");

	    if (skin_font)
	    {
		set_font (self, skin_font);
		xcs_key_set_string (xcs, "bmp", "font", skin_font);
	    }
	}
    }
    else
    {
	skin_font = xcs_key_fetch (xcs, "bmp", "font-save");
	if (skin_font)
	{
	    set_font (self, skin_font);
	    xcs_key_set_string (xcs, "bmp", "font", skin_font);
	}
    }

}

static void
skin_view_update (BmpPreferences *self)
{
    GtkWidget *dialog_preferences;
    GtkTreeView *treeview = GTK_TREE_VIEW(self->private->skin_view);
    GtkTreeSelection *selection = NULL;
    GtkListStore *store;
    GtkTreeIter iter, iter_current_skin;
    GtkTreePath *path;
    GdkPixbuf *thumbnail;
    const gchar *current_skin;
    gchar *name, *basename = NULL, *current_skin_basename = NULL;
    GList *entry;

    dialog_preferences = glade_xml_get_widget_warn (self->private->xml, "dialog_preferences");

    current_skin = xcs_key_fetch (xcs, "bmp", "skin");
    if (current_skin)
      {
	basename = g_path_get_basename (current_skin);
	current_skin_basename = archive_basename (basename);
	if (!current_skin_basename &&
	    g_file_test (current_skin, G_FILE_TEST_IS_DIR))
	  current_skin_basename = basename;
      }

    bmp_window_set_busy (dialog_preferences);
    gtk_notebook_set_current_page (GTK_NOTEBOOK(self->private->skin_view_notebook), 1);

    store = GTK_LIST_STORE(gtk_tree_view_get_model(treeview));
    gtk_list_store_clear(store);

    update_skinlist(self);

    for (entry = self->private->skinlist; entry; entry = g_list_next(entry)) {
        thumbnail = skin_get_thumbnail(SKIN_NODE(entry->data)->path);

        if (!thumbnail)
            continue;

        name = SKIN_NODE(entry->data)->name;

        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           SKIN_VIEW_COL_PREVIEW, thumbnail,
                           SKIN_VIEW_COL_NAME, name,
			   SKIN_VIEW_COL_PATH, SKIN_NODE(entry->data)->path, -1);
        g_object_unref(thumbnail);
	attr_row_map_mapping_insert (map_skins, SKIN_NODE(entry->data)->path, GTK_TREE_MODEL(store), &iter);  

        if ( current_skin_basename && (!strcmp(current_skin_basename,
                    name))) {
            iter_current_skin = iter;
        }

        while (gtk_events_pending())
            gtk_main_iteration();
    }

    selection = gtk_tree_view_get_selection(treeview);
    gtk_tree_selection_select_iter(selection, &iter_current_skin);

    path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter_current_skin);
    gtk_tree_view_scroll_to_cell(treeview, path, NULL, TRUE, 0.5, 0.5);
    gtk_tree_path_free(path);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(self->private->skin_view_notebook), 0);
    bmp_window_set_idle (dialog_preferences);
}

/* Skins */
static gchar *
get_thumbnail_filename (const gchar * path)
{
    gchar *basename, *pngname, *thumbname;

    g_return_val_if_fail(path != NULL, NULL);

    basename = g_path_get_basename(path);
    pngname = g_strconcat(basename, ".png", NULL);

    thumbname = g_build_filename(BMP_PATH_SKIN_THUMB_DIR,
                                 pngname, NULL);

    g_free(basename);
    g_free(pngname);

    return thumbname;
}


static GdkPixbuf *
skin_get_preview (const gchar * path)
{
    GdkPixbuf *preview = NULL;
    gchar *dec_path, *preview_path;
    gboolean is_archive = FALSE;

    if (file_is_archive(path))
      {
        if (!(dec_path = archive_decompress(path)))
            return NULL;

        is_archive = TRUE;
      }
    else
      {
        dec_path = g_strdup(path);
      }

    preview_path = find_file_recursively (dec_path, "main.bmp", 1);

    if (preview_path)
      {
        preview = gdk_pixbuf_new_from_file (preview_path, NULL);
        g_free(preview_path);
      }
    else
      {
	// g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to find main.bmp in path '%s'", G_STRLOC, dec_path);
      }

    if (is_archive)
        del_directory(dec_path);

    g_free(dec_path);

    return preview;
}


static GdkPixbuf *
skin_get_thumbnail (const gchar * path)
{
    GdkPixbuf *scaled = NULL;
    GdkPixbuf *preview;
    gchar *thumbname;

    g_return_val_if_fail (path != NULL, NULL);

    if (g_str_has_suffix (path, BMP_SKIN_THUMB_DIR_BASENAME))
        return NULL;

    thumbname = get_thumbnail_filename (path);

    if (g_file_test(thumbname, G_FILE_TEST_EXISTS))
      {
        scaled = gdk_pixbuf_new_from_file (thumbname, NULL);
        g_free(thumbname);
        return scaled;
    }

    if (!(preview = skin_get_preview(path)))
      {
        g_free(thumbname);
        return NULL;
      }

    scaled = gdk_pixbuf_scale_simple(preview,
                                     THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT,
                                     GDK_INTERP_BILINEAR);
    g_object_unref(preview);

    gdk_pixbuf_save(scaled, thumbname, "png", NULL, NULL);
    g_free(thumbname);

    return scaled;
}

static void
skinlist_add (BmpPreferences *self, const gchar * filename)
{
    SkinNode *node;
    gchar *basename;

    g_return_if_fail(filename != NULL);

    node = g_new0(SkinNode, 1);
    node->path = g_strdup(filename);

    basename = g_path_get_basename(filename);

    if (file_is_archive(filename))
      {
        node->name = archive_basename(basename);
        g_free(basename);
      }
    else
      {
        node->name = basename;
      }


    self->private->skinlist = g_list_prepend(self->private->skinlist, node);
}

static gboolean
scan_skindir_func(const gchar * path, const gchar * basename, BmpPreferences *self)
{

    if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
        if (file_is_archive(path)) {
            skinlist_add(self, path);
        }
    }
    else if (g_file_test(path, G_FILE_TEST_IS_DIR)) {
        skinlist_add(self, path);
    }

    return FALSE;
}

static void
scan_skindir(BmpPreferences *self, const gchar * path)
{
    GError *error = NULL;

    g_return_if_fail(path != NULL);

    if (path[0] == '.')
        return;

    if (!dir_foreach(path, (DirForeachFunc)scan_skindir_func, self, &error)) {
        g_warning("Failed to open directory (%s): %s", path, error->message);
        g_error_free(error);
        return;
    }
}

#if 0
static gint
skinlist_compare_func(gconstpointer a, gconstpointer b)
{
    g_return_val_if_fail(a != NULL && SKIN_NODE(a)->name != NULL, 1);
    g_return_val_if_fail(b != NULL && SKIN_NODE(b)->name != NULL, 1);
    return g_ascii_strcasecmp(SKIN_NODE(a)->name, SKIN_NODE(b)->name);
}
#endif

static void
skin_free_func(gpointer data)
{
    g_return_if_fail(data != NULL);
    g_free(SKIN_NODE(data)->name);
    g_free(SKIN_NODE(data)->path);
    g_free(data);
}

static void
skinlist_clear(BmpPreferences *self)
{
    if (!self->private->skinlist)
        return;

    g_list_foreach(self->private->skinlist, (GFunc) skin_free_func, NULL);
    g_list_free(self->private->skinlist);
    self->private->skinlist = NULL;
}

void
update_skinlist(BmpPreferences *self)
{
    skinlist_clear(self);

    scan_skindir(self, DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME);
    scan_skindir(self, BMP_PATH_USER_SKIN_DIR);

    g_assert(self->private->skinlist != NULL);
}

static void
skin_view_on_cursor_changed(GtkTreeView * treeview,
                            BmpPreferences *self)
{
    GtkTreeModel *model;
    GtkTreeSelection *selection;
    GtkTreeIter iter;
    const gchar *aux;
    gchar *path;

    selection = gtk_tree_view_get_selection(treeview);
    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, SKIN_VIEW_COL_PATH, &path, -1);

    aux = xcs_key_fetch (xcs, "bmp", "skin");
    if(!strcmp(aux, path))
	return;

    /* This will invoke the signal handler in bmp_ui.c for the configuration's skin path value */
    xcs_key_set_string (xcs, "bmp", "skin", path);

    /* Now the font-per-skin handling */
    if (xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin"))
    {
	    const gchar *skin_font;
	    gchar *basename, *basename_archive;

#ifndef USE_GCONF
	    xcs_domain_new (xcs, "skin-fonts");
#endif

	    basename = g_path_get_basename (path);
	    basename_archive = archive_basename (basename);

	    skin_font = xcs_key_fetch(xcs, "skin-fonts", basename_archive);

	    if (skin_font)
	    {
		set_font (self, skin_font);
		xcs_key_set_string (xcs, "bmp", "font", skin_font);
	    }
	    else
	    {
		skin_font = xcs_key_fetch (xcs, "bmp", "font-save");
		if (skin_font)
		{
		    set_font (self, skin_font);
		    xcs_key_set_string (xcs, "bmp", "font", skin_font);
		}
	    }
	    g_free (basename);
	    g_free (basename_archive);
    }

}

static gint
skin_list_sort_func (GtkTreeModel *model,
                     GtkTreeIter *iter_a,
                     GtkTreeIter *iter_b,
                     gpointer user_data)
{
    gchar *string_a, *string_b;
    gint cmp_val;

    gtk_tree_model_get(model, iter_a, SKIN_VIEW_COL_NAME, &string_a, -1);
    gtk_tree_model_get(model, iter_b, SKIN_VIEW_COL_NAME, &string_b, -1);

    cmp_val = g_ascii_strcasecmp(string_a, string_b);

    return cmp_val;
}

static void
skin_added_func (const gchar *path, BmpPreferences *self)
{
    GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(self->private->skin_view)));
    GdkPixbuf *thumbnail;
    GtkTreeIter iter;
    gchar *name, *basename;

    thumbnail = skin_get_thumbnail(path);
    if (!thumbnail)
	    return;
    
    basename = g_path_get_basename(path);

    if (file_is_archive(path))
      {
        name = archive_basename(basename);
        g_free(basename);
      }
    else
      {
        name = basename;
      }

    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter,
	    SKIN_VIEW_COL_PREVIEW, thumbnail,
	    SKIN_VIEW_COL_NAME, name,
	    SKIN_VIEW_COL_PATH, path, -1);
    g_object_unref(thumbnail);
    attr_row_map_mapping_insert (map_skins, path, GTK_TREE_MODEL(store), &iter);  

#if 0
    g_free (name);
    g_free (path);
#endif
}

static void
skin_removed_func (const gchar *path, BmpPreferences *self)
{
	GtkTreeModel *store = gtk_tree_view_get_model(GTK_TREE_VIEW(self->private->skin_view));
	GtkTreeIter iter;
	gchar *name;
	gboolean loop = TRUE;

	if (!gtk_tree_model_get_iter_first (store, &iter)) return;

	while (loop)
	  {
		gtk_tree_model_get (store, &iter, SKIN_VIEW_COL_PATH, &name, -1);
		if (!strcmp(path,name))
		  {
			gchar *path;

		        gtk_tree_model_get (GTK_TREE_MODEL(store), &iter, SKIN_VIEW_COL_PATH, &path, -1);
			attr_row_map_mapping_remove (map_skins, path);
			gtk_list_store_remove (GTK_LIST_STORE(store), &iter);
#if 0
			g_free (path);
			g_free (name);
#endif
			loop = FALSE;
		  }
		else
		  {
			if (!gtk_tree_model_iter_next (store, &iter)) loop = FALSE;
		  }
	  }
#if 0
       g_free(path);
#endif
}

static void
column_toggled  (GtkCellRendererToggle *cell,
                 gchar                 *path_str,
                 BmpPreferences        *self)
{
    BmpWindowPlaylist *ui_playlist;
    GtkTreePath	      *path;
    GtkTreeIter	       iter;
    gboolean	       new_state;
    gint	       column;

    ui_playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);

    path = gtk_tree_path_new_from_string (path_str); 
    gtk_tree_model_get_iter (GTK_TREE_MODEL(self->private->columns_store), &iter, path);
    column = gtk_tree_path_get_indices (path)[0];
    gtk_tree_path_free (path);

    new_state = gtk_cell_renderer_toggle_get_active (cell) ? FALSE : TRUE;
    gtk_list_store_set (self->private->columns_store, &iter, 0, new_state, -1);
    bmp_window_playlist_column_set_active (ui_playlist, column, new_state); 
}

static void
columns_view_setup (BmpPreferences *self)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(self->private->columns_view);
    GtkTreeViewColumn *column;
    GtkCellRenderer *renderer;
    GtkTreeIter iter;
    gint n;

    self->private->columns_store = gtk_list_store_new(3, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_INT);

    for (n = 0; n < vcolumns_get_n_items (); n++)
      {
	  gchar	    *key_name;
	  gboolean   visible;

	  key_name = g_strdup_printf ("column_%d_visible", n);
	  visible = xcs_key_fetch_boolean (xcs, "tracklist", key_name);
	  g_free (key_name);
 
	  gtk_list_store_append (self->private->columns_store, &iter);
	  gtk_list_store_set (self->private->columns_store, &iter, 0, visible, 1, bmp_metadata_get_title_static (vcolumns_get_item_nth(n)), -1);
      }

    column = gtk_tree_view_column_new ();
    renderer = gtk_cell_renderer_toggle_new ();
    gtk_tree_view_column_pack_start (column, renderer, FALSE);
    gtk_tree_view_column_set_attributes (column, renderer, "active", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), GTK_TREE_VIEW_COLUMN(column));

    signal_handler_pool_add_connection (self->private->pool,
					renderer,	
					"toggled",
					G_CALLBACK(column_toggled),
					self,	
					0);	


    column = gtk_tree_view_column_new ();
    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_set_attributes(column, renderer, "text", 1, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), GTK_TREE_VIEW_COLUMN(column));

    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(self->private->columns_store));
}


static void
skin_view_setup (BmpPreferences *self)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(self->private->skin_view);
    GtkListStore *store;
    GtkTreeViewColumn *column;
    GtkCellRenderer *renderer;
    GtkTreeSelection *selection;

    gtk_widget_show_all(GTK_WIDGET(treeview));

    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);

    store = gtk_list_store_new(SKIN_VIEW_N_COLS, GDK_TYPE_PIXBUF,
                               G_TYPE_STRING, G_TYPE_STRING);
    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    map_skins = attr_row_map_new ();

    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
    gtk_tree_view_column_set_spacing(column, 16);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview),
                                GTK_TREE_VIEW_COLUMN(column));

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_set_attributes(column, renderer, "pixbuf",
                                        SKIN_VIEW_COL_PREVIEW, NULL);
    g_object_set (G_OBJECT(renderer),"xpad",6,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_set_attributes(column, renderer, "text",
                                        SKIN_VIEW_COL_NAME, NULL);

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE(store),
                                     SKIN_VIEW_COL_NAME,
                                     skin_list_sort_func,
                                     NULL,
                                     NULL);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(store), SKIN_VIEW_COL_NAME, GTK_SORT_ASCENDING);

    signal_handler_pool_add_connection (self->private->pool,
					treeview,	
					"cursor-changed",
					G_CALLBACK(skin_view_on_cursor_changed),
					self,	
					0);	
}

/* Plugins */
static void
on_button_plugins_show_all  (GtkWidget   *button,
			     gpointer	  data)
{
    BmpPreferences  *self = BMP_PREFERENCES (data);
    GtkWidget	    *notebook;

    notebook = glade_xml_get_widget_warn (self->private->xml, "plugins_notebook_main");

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook), 0);
    gtk_widget_set_sensitive (button, FALSE);
}

static void 
on_plugins_icon_view_item_activated (GtkIconView *iconview,
                                     GtkTreePath *path,
                                     gpointer	  data)
{
    BmpPreferences  *self = BMP_PREFERENCES (data);
    GtkWidget	    *button, *notebook;

    notebook = glade_xml_get_widget_warn (self->private->xml, "plugins_notebook_main");
    button = glade_xml_get_widget_warn (self->private->xml, "b_plugins_show_all");

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook),
				   gtk_tree_path_get_indices (path)[0]+1); 
    gtk_widget_set_sensitive (button, TRUE);
}

typedef struct _PluginActivationTuple PluginActivationTuple;
struct _PluginActivationTuple {
    BmpPreferences    *self;
    PluginType	       type;
};

static void 
on_plugin_sub_icon_view_item_activated (GtkIconView	      *iconview,
					GtkTreePath	      *path,
					PluginActivationTuple *tuple) 
{
  GtkTreeModel *model = gtk_icon_view_get_model (iconview);
  GtkTreeIter   iter;
  gchar	       *name;
  
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (model, &iter, 2, &name, -1);

  bmp_plugin_host_gui_show (tuple->self->private->host,
				     tuple->type, 
				     name); 
}

static void
append_to_list (gpointer _k, gpointer _v, GList **list)
{
    if (BMP_IS_PLUG_PROPERTY(_v))
	  *list = g_list_append (*list, _k); 
}

static void
icon_view_set_bmpx_defaults (GtkIconView *icon_view)
{
    gtk_icon_view_set_text_column   (icon_view, 0);
    gtk_icon_view_set_pixbuf_column (icon_view, 1);
    gtk_icon_view_set_item_width    (icon_view, 128);
    gtk_icon_view_set_orientation   (icon_view, GTK_ORIENTATION_VERTICAL);
    gtk_icon_view_set_columns	    (icon_view, -1);
}

static void
on_plugins_iconview_realize (GtkIconView *icon_view,
			     gpointer	  data)
{
    BmpPreferences	    *self = BMP_PREFERENCES (data);
    GtkListStore	    *icon_store, *icon_store_sub; 
    GtkWidget		    *icon_view_sub;
    GdkPixbuf		    *pb;
    GtkTreeIter		     iter;
    PluginActivationTuple   *tuple;
    GList		    *plugins = NULL;
    gint		     n;

    g_return_if_fail (GTK_IS_ICON_VIEW (icon_view));  

    icon_store = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_INT);
    for (n = 0; n < BMP_N_PLUGIN_TYPES; n++)
      {
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 
	gtk_list_store_append (icon_store, &iter);	
	gtk_list_store_set (icon_store, &iter, 0, plugin_type_get_desc (n), 1, pb, 2, n, -1);

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (icon_view, GTK_TREE_MODEL(icon_store));
    icon_view_set_bmpx_defaults (icon_view);

    signal_handler_pool_add_connection (self->private->pool,
					icon_view,	
					"item-activated",
					G_CALLBACK(on_plugins_icon_view_item_activated),
					self,	
					0);	

    signal_handler_pool_add_connection (self->private->pool,
					glade_xml_get_widget_warn (self->private->xml, "b_plugins_show_all"),
					"clicked",
					G_CALLBACK(on_button_plugins_show_all),
					self,	
					0);	

    /* Now set up the sub-icon-views */
    //Transport
    icon_view_sub = glade_xml_get_widget_warn (self->private->xml, "iconview_transport");    
    icon_store_sub = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING);
    g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_TRANSPORT], (GHFunc)append_to_list, &plugins);

    for ( ; plugins ; plugins = plugins->next)
      {
	GObject	     *object;
	const gchar **metadata;
    
	object = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_TRANSPORT], plugins->data);
	metadata = bmp_plug_metadata_get (BMP_PLUG_METADATA (object));
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 

	gtk_list_store_append (icon_store_sub, &iter);	
	gtk_list_store_set (icon_store_sub, &iter, 0, metadata[METADATA_FIELD_NAME], 1, pb, 2, (gchar*)plugins->data, -1);

	bmp_plugin_host_create_gui_for_plugin (self->private->host,
							BMP_PLUGIN_TYPE_TRANSPORT,	
							(const gchar *)plugins->data);	

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (GTK_ICON_VIEW(icon_view_sub), GTK_TREE_MODEL(icon_store_sub));
    icon_view_set_bmpx_defaults (GTK_ICON_VIEW(icon_view_sub));

    tuple = g_new0 (PluginActivationTuple,1);
    tuple->self = self;
    tuple->type = BMP_PLUGIN_TYPE_TRANSPORT;

    signal_handler_pool_add_connection (self->private->pool,
					icon_view_sub,	
					"item-activated",
					G_CALLBACK(on_plugin_sub_icon_view_item_activated),
					tuple,	
					0);	

    //Flow
    plugins = NULL;
    icon_view_sub = glade_xml_get_widget_warn (self->private->xml, "iconview_flow");    
    icon_store_sub = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING);
    g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_FLOW], (GHFunc)append_to_list, &plugins);

    for ( ; plugins ; plugins = plugins->next)
      {
	GObject	     *object;
	const gchar **metadata;
    
	object = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_FLOW], plugins->data);
	metadata = bmp_plug_metadata_get (BMP_PLUG_METADATA (object));
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 

	gtk_list_store_append (icon_store_sub, &iter);	
	gtk_list_store_set (icon_store_sub, &iter, 0, metadata[METADATA_FIELD_NAME], 1, pb, 2, (gchar*)plugins->data, -1);

	bmp_plugin_host_create_gui_for_plugin (self->private->host,
							BMP_PLUGIN_TYPE_FLOW,	
							(const gchar *)plugins->data);	

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (GTK_ICON_VIEW(icon_view_sub), GTK_TREE_MODEL(icon_store_sub));
    icon_view_set_bmpx_defaults (GTK_ICON_VIEW(icon_view_sub));

    tuple = g_new0 (PluginActivationTuple,1);
    tuple->self = self;
    tuple->type = BMP_PLUGIN_TYPE_FLOW;

    signal_handler_pool_add_connection (self->private->pool,
					icon_view_sub,	
					"item-activated",
					G_CALLBACK(on_plugin_sub_icon_view_item_activated),
					tuple,	
					0);	
}

static void
on_category_view_realize (GtkTreeView	  *treeview,
                          BmpPreferences  *self)
{
    GtkListStore      *store;
    GtkCellRenderer   *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection  *selection;
    GtkWidget	      *notebook;
    GdkPixbuf	      *img;
    GtkTreeIter	       iter;
    gint	       i;

    notebook = glade_xml_get_widget(self->private->xml, "category_notebook");
    column = gtk_tree_view_column_new();
    gtk_tree_view_append_column(treeview, column);
    gtk_tree_view_column_set_spacing(column, 2);

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_attributes(column, renderer, "pixbuf", CATEGORY_VIEW_COL_ICON, NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_attributes(column, renderer, "text", CATEGORY_VIEW_COL_NAME, NULL);


    store = gtk_list_store_new (CATEGORY_VIEW_N_COLS,
                                GDK_TYPE_PIXBUF, 
				G_TYPE_STRING, 
				G_TYPE_INT);

    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    for (i = 0; i < n_categories; i++) {
        img = gdk_pixbuf_new_from_file(categories[i].icon_path, NULL);
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           CATEGORY_VIEW_COL_ICON,
			   img,
                           CATEGORY_VIEW_COL_NAME,
                           gettext(categories[i].name),
			   CATEGORY_VIEW_COL_ID,
                           categories[i].id,
			   -1);
        g_object_unref(img);
    }

    gtk_tree_view_set_headers_visible (treeview, FALSE);

    selection = gtk_tree_view_get_selection(treeview);

    signal_handler_pool_add_connection (self->private->pool,
					selection,	
					"changed",
					G_CALLBACK(change_category),
					notebook,	
					G_CONNECT_SWAPPED);	
}

/* FIXME: complete the map */
FUNC_MAP_BEGIN(prefswin_func_map)
    FUNC_MAP_ENTRY(on_custom_cursors_realize)
    FUNC_MAP_ENTRY(on_custom_cursors_toggled)
    FUNC_MAP_ENTRY(on_font_per_skin_realize)
    FUNC_MAP_ENTRY(on_font_per_skin_toggled)
    FUNC_MAP_ENTRY(on_display_tooltips_realize)
    FUNC_MAP_ENTRY(on_display_numbers_realize)
    FUNC_MAP_ENTRY(on_font_button_font_set)
    FUNC_MAP_ENTRY(on_font_button_realize)
    FUNC_MAP_ENTRY(on_plugins_iconview_realize)
FUNC_MAP_END

#if 0
static void
on_eq_value_changed (GtkWidget *widget, gpointer user_data)
{
    gdouble value;
    gint value_int;

    g_return_if_fail (GTK_IS_WIDGET(widget));

    value = gtk_range_get_value (GTK_RANGE(widget));
    value_int = (gint)value;

    bmp_system_control_eq_set (bmp_system_control, GPOINTER_TO_INT(user_data), value_int, NULL);
}
#endif

static void
lastfm_enable_changed	    (BmpPreferences *self,
			     gpointer	     data) 
{
    GtkWidget *checkbutton;
    gboolean   active;

    active = xcs_key_fetch_boolean (xcs, "lastfm", "enable");

    checkbutton = glade_xml_get_widget_warn (self->private->xml, "last_fm_enable");

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (checkbutton)) != active)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkbutton), active);
       
}

void
window_preferences_create   (BmpPreferences *self)
{
    BmpTooltips		*tooltips;
    GtkWidget		*widget;
    const gchar		*glade_file = DATA_DIR "/glade/dialog_preferences.glade";
    gchar		*aux;
    gint		 i;

    /* o Initialize tooltips */
    tooltips = bmp_tooltips_new ();

    RM_ITEM_NEW("tooltips", "tooltips-preferences", tooltips, NULL);

    /* o Load the interface */
    self->private->xml = glade_xml_new_or_die(Q_("Preferences Window"), glade_file, NULL, NULL);

    /* o Connect the signals in the interface */
    glade_xml_signal_autoconnect_map (self->private->xml, prefswin_func_map, self);

    self->window = glade_xml_get_widget (self->private->xml, "dialog_preferences");

    /* o Set icon to window */
    bmp_window_set_icon_list (GTK_WIDGET(self->window), "player");

    /* o Create Category View */
    widget = glade_xml_get_widget(self->private->xml, "category_view");

    signal_handler_pool_add_connection (self->private->pool,
					widget,	
					"realize",
					G_CALLBACK(on_category_view_realize),
					self,	
					G_CONNECT_AFTER);	

    /*  o Skin View */
    self->private->skin_view = glade_xml_get_widget (self->private->xml, "skin_view");
    self->private->skin_view_notebook = glade_xml_get_widget (self->private->xml, "skins_notebook");
    skin_view_setup (self);

    /*  o Columns Treeview */
    self->private->columns_view = glade_xml_get_widget (self->private->xml, "columns_view");
    columns_view_setup (self);
 
    /* o Titlestring */
    widget = glade_xml_get_widget(self->private->xml, "titlestring_entry");
    gtk_entry_set_text (GTK_ENTRY(GTK_BIN(widget)->child), xcs_key_fetch(xcs,"bmp","titlestring-format"));

    xcs_bind_object (xcs,
		     G_OBJECT(widget),
		     titlestring_binding,
		     "changed",
		     "bmp",
		     "titlestring-format"); 

    /* o Images */
    widget = glade_xml_get_widget(self->private->xml, "image_building");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget(self->private->xml, "last_fm_header_logo");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/lastfm-logo.png");

    widget = glade_xml_get_widget(self->private->xml, "last_fm_poweredby");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/audioscrobbler.png");


#if 0
    /* Connect eq sliders */
    for (i = 0; i < n_eq_sliders; i++)
      {
	widget = glade_xml_get_widget(self->private->xml, eq_sliders[i]);
	g_signal_connect (G_OBJECT(widget), "value-changed", G_CALLBACK(on_eq_value_changed), GINT_TO_POINTER(i));
      }
#endif

    for (i = 0; i < n_window_preferences_tooltips; i++)
      {
	widget = glade_xml_get_widget(self->private->xml, window_preferences_tooltips[i].widget_name);
	bmp_tooltips_set_tip (tooltips,
				widget,
				window_preferences_tooltips[i].stock_id,
				gettext (window_preferences_tooltips[i].tooltip_text),
			        NULL,
				TRUE);
      }

    widget = glade_xml_get_widget(self->private->xml, "label_version");
    gtk_label_set_markup (GTK_LABEL(widget), 
		  "<span size=\"small\" color=\"#606060\">BMPx " RV_LAST_CHANGED_DATE "-R" RV_REVISION " [" BUILD_DATE "] (" BUILD_BUILDUSER ")</span>"); 

    /* Bind widgets/other objects */ 

    /* XXX: Should we pull a DBus and derive the configuration key names from the widget name, a la "custom_cursors" -> "custom-cursors"? */
    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn (self->private->xml, "custom_cursors")),
			    "bmp",
			    "use-custom-cursors");

    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn (self->private->xml, "display_tooltips")),
			    "bmp",
			    "display-tooltips");

    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn (self->private->xml, "display_numbers")),
			    "bmp",
			    "display-tracklist-numbers");

    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn (self->private->xml, "font_per_skin")),
			    "bmp",
			    "font-per-skin");

    xcs_bind_object (xcs,
		     G_OBJECT(glade_xml_get_widget_warn (self->private->xml,"font_button")),
		     fontbutton_binding,
		     "font_set",
		     "bmp",
		     "font"); 

    /* Last.FM stuff */
    widget = glade_xml_get_widget(self->private->xml, "last_fm_enable");
    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(widget),
			    "lastfm",
			    "enable");
    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::lastfm-enable",
					G_CALLBACK(lastfm_enable_changed),
				        self,
					G_CONNECT_SWAPPED);	

    widget = glade_xml_get_widget(self->private->xml, "last_fm_username");
    aux = xcs_key_fetch_string (xcs, "lastfm", "username");
    gtk_entry_set_text (GTK_ENTRY(widget), aux);
    g_free (aux);

    widget = glade_xml_get_widget(self->private->xml, "last_fm_password");
    aux = xcs_key_fetch_string (xcs, "lastfm", "password");
    gtk_entry_set_text (GTK_ENTRY(widget), aux);
    g_free (aux);

    xcs_bind_object (xcs,
		     G_OBJECT(glade_xml_get_widget_warn (self->private->xml,"last_fm_username")),
		     gtkentry_binding,
		     "changed",
		     "lastfm",
		     "username"); 

    xcs_bind_object (xcs,
		     G_OBJECT(glade_xml_get_widget_warn (self->private->xml,"last_fm_password")),
		     gtkentry_binding,
		     "changed",
		     "lastfm",
		     "password"); 

    widget = glade_xml_get_widget(self->private->xml, "last_fm_enable");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widget), xcs_key_fetch_boolean (xcs, "lastfm", "enable"));

    /* Various watches */
    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-display-tracklist-numbers",
					G_CALLBACK(display_tracklist_numbers_changed),
				        self,
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-display-tooltips",
					G_CALLBACK(display_tooltips_changed),
					glade_xml_get_widget_warn (self->private->xml, "display_tooltips"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-font",
					G_CALLBACK(font_changed),
					glade_xml_get_widget_warn (self->private->xml, "font_button"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-use-custom-cursors",
					G_CALLBACK(use_custom_cursors_changed),
					glade_xml_get_widget_warn (self->private->xml, "custom_cursors"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-font-per-skin",
					G_CALLBACK(font_per_skin_changed),
					glade_xml_get_widget_warn (self->private->xml, "font_per_skin"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (self->private->pool,
					xcs,	
					"changed::bmp-skin",
					G_CALLBACK(skin_changed),
					self,	
					G_CONNECT_SWAPPED);	




    gtk_widget_realize (self->window);

    if (xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips"))
	bmp_tooltips_enable (tooltips);
    else
	bmp_tooltips_disable (tooltips);
}

void
bmp_preferences_show (BmpPreferences *self)
{
    gtk_window_set_position (GTK_WINDOW (self->window), GTK_WIN_POS_CENTER_ALWAYS);
    gtk_window_present (GTK_WINDOW(self->window));

    if (!self->private->updated)
    {
        self->private->updated = TRUE;
        skin_view_update(self);
	gtk_widget_show_all(self->private->skin_view);

	self->private->skins_path = g_build_filename(DATA_DIR,G_DIR_SEPARATOR_S,BMP_SKIN_DIR_BASENAME,NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"skins",
				self->private->skins_path,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				self,
				NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"user skins",
				BMP_PATH_USER_SKIN_DIR,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				self,
				NULL);

    }
}
